import re

# from https://github.com/pypa/pip/blob/303fed36c1771de4063063a866776a9103972317/src/pip/_internal/vcs/git.py#L40
# SCP (Secure copy protocol) shorthand. e.g. 'git@example.com:foo/bar.git'
SCP_REGEX = re.compile(
    r"""^
    # Optional user, e.g. 'git@'
    (\w+@)?
    # Server, e.g. 'github.com'.
    ([^/:]+):
    # The server-side path. e.g. 'user/project.git'. Must start with an
    # alphanumeric character so as not to be confusable with a Windows paths
    # like 'C:/foo/bar' or 'C:\foo\bar'.
    (\w[^:]*)
    $""",
    re.VERBOSE,
)


def is_scp_style_url(url: str) -> bool:
    return bool(SCP_REGEX.match(url))



"""DVC re-implementation of fsspec's dedicated async event loop."""

import asyncio
import os
import threading
from typing import Any, Optional

from fsspec.asyn import (  # noqa: F401, pylint:disable=unused-import
    sync,
    sync_wrapper,
)

# dedicated async IO thread
iothread: list[Optional[threading.Thread]] = [None]
# global DVC event loop
default_loop: list[Optional[asyncio.AbstractEventLoop]] = [None]
lock = threading.Lock()


def get_loop() -> asyncio.AbstractEventLoop:
    """Create or return the global DVC event loop."""
    if default_loop[0] is None:
        with lock:
            if default_loop[0] is None:
                default_loop[0] = asyncio.new_event_loop()
                loop = default_loop[0]
                th = threading.Thread(
                    target=loop.run_forever,  # type: ignore[attr-defined]
                    name="dvcIO",
                )
                th.daemon = True
                th.start()
                iothread[0] = th
    assert default_loop[0] is not None
    return default_loop[0]


class BaseAsyncObject:
    def __init__(
        self, loop: Optional[asyncio.AbstractEventLoop] = None, **kwargs: Any
    ) -> None:
        self._loop: asyncio.AbstractEventLoop = loop or get_loop()
        self._pid = os.getpid()

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        # AsyncMixin is not fork-safe
        assert self._pid == os.getpid()
        return self._loop



import asyncio
import os
import subprocess
import sys
from collections.abc import AsyncIterator, Iterator
from typing import Any

import asyncssh
import pygit2
import pytest
from pytest_test_utils import TempDirFactory, TmpDir

from scmrepo.git import Git

TEST_SSH_USER = "user"
TEST_SSH_KEY_PATH = os.path.join(
    os.path.abspath(os.path.dirname(__file__)), f"{TEST_SSH_USER}.key"
)

# pylint: disable=redefined-outer-name


def pytest_addoption(parser):
    parser.addoption(
        "--slow", action="store_true", default=False, help="run slow tests"
    )


def pytest_collection_modifyitems(config, items):
    if config.getoption("--slow"):
        return
    skip_slow = pytest.mark.skip(reason="need --slow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


@pytest.fixture(autouse=True)
def _isolate(tmp_dir_factory: TempDirFactory, monkeypatch: pytest.MonkeyPatch) -> None:
    path = tmp_dir_factory.mktemp("mock")
    home_dir = path / "home"
    home_dir.mkdir()

    if sys.platform == "win32":
        home_drive, home_path = os.path.splitdrive(home_dir)
        monkeypatch.setenv("USERPROFILE", str(home_dir))
        monkeypatch.setenv("HOMEDRIVE", home_drive)
        monkeypatch.setenv("HOMEPATH", home_path)
    else:
        monkeypatch.setenv("HOME", str(home_dir))

    monkeypatch.setenv("GIT_CONFIG_NOSYSTEM", "1")
    contents = b"""
[user]
name=DVC Tester
email=dvctester@example.com
[init]
defaultBranch=master
"""
    (home_dir / ".gitconfig").write_bytes(contents)
    pygit2.settings.search_path[pygit2.GIT_CONFIG_LEVEL_GLOBAL] = str(home_dir)  # type: ignore[attr-defined]


@pytest.fixture
def scm(tmp_dir: TmpDir) -> Iterator[Git]:
    git_ = Git.init(tmp_dir)
    sig = git_.pygit2.default_signature

    assert sig.email == "dvctester@example.com"
    assert sig.name == "DVC Tester"

    yield git_
    git_.close()


backends = ["gitpython", "dulwich", "pygit2"]


@pytest.fixture(params=backends)
def git_backend(request) -> str:
    marker = request.node.get_closest_marker("skip_git_backend")
    to_skip = marker.args if marker else []

    backend = request.param
    if backend in to_skip:
        pytest.skip()
    return backend


@pytest.fixture
def git(tmp_dir: TmpDir, git_backend: str) -> Iterator[Git]:
    git_ = Git(tmp_dir, backends=[git_backend])
    yield git_
    git_.close()


@pytest.fixture
def remote_git_dir(tmp_dir_factory: TempDirFactory):
    git_dir = tmp_dir_factory.mktemp("git-remote")
    remote_git = Git.init(git_dir)
    remote_git.close()
    return git_dir


@pytest.fixture(scope="session")
def docker(request: pytest.FixtureRequest):
    for cmd in [("docker", "ps"), ("docker", "compose", "version")]:
        try:
            subprocess.check_call(
                cmd,
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
            )
        except (subprocess.CalledProcessError, OSError):
            pytest.skip(f"no {cmd[0]} installed")

    if "CI" in os.environ and os.name == "nt":
        pytest.skip("disabled for Windows on Github Actions")

    pytest.importorskip("pytest_docker")
    return request.getfixturevalue("docker_services")


@pytest.fixture
def ssh_conn_info(
    docker,  # pylint: disable=unused-argument
) -> dict[str, Any]:
    conn_info = {
        "host": "127.0.0.1",
        "port": docker.port_for("git-server", 2222),
        "client_keys": TEST_SSH_KEY_PATH,
        "known_hosts": None,
        "username": TEST_SSH_USER,
    }

    async def _check() -> bool:
        try:
            async with asyncssh.connect(**conn_info) as conn:
                result = await conn.run("git --version")
                assert result.returncode == 0
                async with conn.start_sftp_client() as sftp:
                    assert await sftp.exists("/")
        except Exception:  # noqa: BLE001 # pylint: disable=broad-except
            return False
        return True

    def check() -> bool:
        return asyncio.run(_check())

    docker.wait_until_responsive(timeout=30.0, pause=1, check=check)
    return conn_info


@pytest.fixture
async def ssh_connection(
    ssh_conn_info: dict[str, Any],
) -> AsyncIterator[asyncssh.connection.SSHClientConnection]:
    async with asyncssh.connect(**ssh_conn_info) as conn:
        yield conn


@pytest.fixture
async def sftp(
    ssh_connection: asyncssh.connection.SSHClientConnection,
) -> AsyncIterator[asyncssh.SFTPClient]:
    async with ssh_connection.start_sftp_client() as sftp:
        yield sftp



import errno
import os
import posixpath
from typing import TYPE_CHECKING, Any, BinaryIO, Callable, Optional

from fsspec.spec import AbstractFileSystem

if TYPE_CHECKING:
    from io import BytesIO

    from scmrepo.git import Git
    from scmrepo.git.objects import GitTrie


def bytesio_len(obj: "BytesIO") -> Optional[int]:
    try:
        offset = obj.tell()
        length = obj.seek(0, os.SEEK_END)
        obj.seek(offset)
    except (AttributeError, OSError):
        return None
    return length


class GitFileSystem(AbstractFileSystem):
    # pylint: disable=abstract-method
    cachable = False
    root_marker = "/"

    def __init__(
        self,
        path: Optional[str] = None,
        rev: Optional[str] = None,
        scm: Optional["Git"] = None,
        trie: Optional["GitTrie"] = None,
        rev_resolver: Optional[Callable[["Git", str], str]] = None,
        **kwargs,
    ):
        from scmrepo.git import Git
        from scmrepo.git.objects import GitTrie

        super().__init__(**kwargs)
        if not trie:
            scm = scm or Git(path)
            resolver = rev_resolver or Git.resolve_rev
            resolved = resolver(scm, rev or "HEAD")
            tree_obj = scm.pygit2.get_tree_obj(rev=resolved)
            trie = GitTrie(tree_obj, resolved)

        self.trie = trie
        self.rev = self.trie.rev
        self._cwd = self.root_marker

    def getcwd(self):
        return self._cwd

    def chdir(self, path):
        self._cwd = path

    @classmethod
    def join(cls, *parts):
        return posixpath.join(*parts)

    @classmethod
    def split(cls, path):
        return posixpath.split(path)

    def normpath(self, path):
        return posixpath.normpath(path)

    @classmethod
    def isabs(cls, path):
        return posixpath.isabs(path)

    def abspath(self, path):
        if not self.isabs(path):
            path = self.join(self.getcwd(), path)
        return self.normpath(path)

    @classmethod
    def commonprefix(cls, path):
        return posixpath.commonprefix(path)

    @classmethod
    def parts(cls, path):
        ret = []
        while True:
            path, part = cls.split(path)

            if part:
                ret.append(part)
                continue

            if path:
                ret.append(path)

            break

        ret.reverse()

        return tuple(ret)

    @classmethod
    def parent(cls, path):
        return posixpath.dirname(path)

    @classmethod
    def dirname(cls, path):
        return cls.parent(path)

    @classmethod
    def parents(cls, path):
        parts = cls.parts(path)
        return tuple(
            cls.join(*parts[:length]) for length in range(len(parts) - 1, 0, -1)
        )

    @classmethod
    def name(cls, path):
        return cls.parts(path)[-1]

    @classmethod
    def suffix(cls, path):
        name = cls.name(path)
        _, dot, suffix = name.partition(".")
        return dot + suffix

    @classmethod
    def with_name(cls, path, name):
        parts = list(cls.parts(path))
        parts[-1] = name
        return cls.join(*parts)

    @classmethod
    def with_suffix(cls, path, suffix):
        parts = list(cls.parts(path))
        real_path, _, _ = parts[-1].partition(".")
        parts[-1] = real_path + suffix
        return cls.join(*parts)

    @classmethod
    def isin(cls, left, right):
        left_parts = cls.parts(left)
        right_parts = cls.parts(right)
        left_len = len(left_parts)
        right_len = len(right_parts)
        return left_len > right_len and left_parts[:right_len] == right_parts

    @classmethod
    def isin_or_eq(cls, left, right):
        return left == right or cls.isin(left, right)

    @classmethod
    def overlaps(cls, left, right):
        # pylint: disable=arguments-out-of-order
        return cls.isin_or_eq(left, right) or cls.isin(right, left)

    def relpath(self, path, start=None):
        if start is None:
            start = "."
        return posixpath.relpath(self.abspath(path), start=self.abspath(start))

    def relparts(self, path, start=None):
        return self.parts(self.relpath(path, start=start))

    @classmethod
    def as_posix(cls, path):
        return path

    def _get_key(self, path: str) -> tuple[str, ...]:
        path = self.abspath(path)
        if path == self.root_marker:
            return ()
        relparts = path.split(self.sep)
        if relparts and relparts[0] in (".", ""):
            relparts = relparts[1:]
        return tuple(relparts)

    def _open(
        self,
        path: str,
        mode: str = "rb",
        block_size: Optional[int] = None,
        autocommit: bool = True,
        cache_options: Optional[dict] = None,
        raw: bool = False,
        **kwargs: Any,
    ) -> BinaryIO:
        if mode != "rb":
            raise NotImplementedError

        key = self._get_key(path)
        try:
            obj = self.trie.open(key, mode=mode, raw=raw)
            obj.size = bytesio_len(obj)
            return obj
        except KeyError as exc:
            msg = os.strerror(errno.ENOENT) + f" in branch '{self.rev}'"
            raise FileNotFoundError(errno.ENOENT, msg, path) from exc
        except IsADirectoryError as exc:
            raise IsADirectoryError(
                errno.EISDIR, os.strerror(errno.EISDIR), path
            ) from exc

    def info(self, path: str, **kwargs: Any) -> dict[str, Any]:
        key = self._get_key(path)
        try:
            # NOTE: to avoid wasting time computing object size, trie.info
            # will return a LazyDict instance, that will compute compute size
            # only when it is accessed.
            ret = self.trie.info(key)
            ret["name"] = path
            return ret
        except KeyError as exc:
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), path
            ) from exc

    def exists(self, path: str, **kwargs: Any) -> bool:
        key = self._get_key(path)
        return self.trie.exists(key)

    def checksum(self, path: str) -> str:
        return self.info(path)["sha"]

    def ls(self, path, detail=True, **kwargs):
        info = self.info(path)
        if info["type"] != "directory":
            return [info] if detail else [path]

        key = self._get_key(path)
        try:
            names = self.trie.ls(key)
        except KeyError as exc:
            raise FileNotFoundError from exc

        paths = [posixpath.join(path, name) if path else name for name in names]

        if not detail:
            return paths

        return [self.info(_path) for _path in paths]



import io
import logging
from typing import TYPE_CHECKING, Callable, Optional

from pygit2 import GIT_FILTER_CLEAN, Filter, Passthrough  # type: ignore[attr-defined]

if TYPE_CHECKING:
    from pygit2 import FilterSource  # type: ignore[attr-defined]

logger = logging.getLogger(__name__)


class LFSFilter(Filter):
    attributes = "filter=*"

    def __init__(self, *args, **kwargs) -> None:
        self._smudge_buf: Optional[io.BytesIO] = None
        self._smudge_root: Optional[str] = None

    def check(self, src: "FilterSource", attr_values: list[Optional[str]]):
        if attr_values[0] == "lfs" and src.mode != GIT_FILTER_CLEAN:  # type: ignore[attr-defined]
            self._smudge_buf = io.BytesIO()
            self._smudge_root = src.repo.workdir or src.repo.path  # type: ignore[attr-defined]
            return
        raise Passthrough

    def write(
        self, data: bytes, src: "FilterSource", write_next: Callable[[bytes], None]
    ):
        if src.mode == GIT_FILTER_CLEAN:  # type: ignore[attr-defined]
            write_next(data)
            return
        if self._smudge_buf is None:
            self._smudge_buf = io.BytesIO()
        if self._smudge_root is None:
            self._smudge_root = src.repo.workdir or src.repo.path  # type: ignore[attr-defined]
        self._smudge_buf.write(data)

    def close(self, write_next: Callable[[bytes], None]):
        if self._smudge_buf is not None:
            assert self._smudge_root
            self._smudge(write_next)

    def _smudge(self, write_next: Callable[[bytes], None]):
        from scmrepo.exceptions import InvalidRemote
        from scmrepo.git import Git
        from scmrepo.git.lfs import smudge
        from scmrepo.git.lfs.fetch import get_fetch_url

        assert self._smudge_buf is not None
        self._smudge_buf.seek(0)
        with Git(self._smudge_root) as scm:
            try:
                url = get_fetch_url(scm)
            except InvalidRemote:
                url = None
            fobj = smudge(scm.lfs_storage, self._smudge_buf, url=url)
            with fobj:
                data = fobj.read(io.DEFAULT_BUFFER_SIZE)
                try:
                    while data:
                        write_next(data)
                        data = fobj.read(io.DEFAULT_BUFFER_SIZE)
                except KeyboardInterrupt:
                    return



import os
from collections.abc import Iterator, MutableMapping
from typing import Callable, TypeVar, Union

K = TypeVar("K")
V = TypeVar("V")


class LazyDict(MutableMapping[K, V]):
    def __init__(self, values: dict[K, Union[V, Callable[[], V]]]):
        self._values = values

    def __getitem__(self, item):
        value = self._values[item]
        if callable(value):
            value = value()
            self._values[item] = value
        return value

    def __setitem__(self, key: K, value: Union[V, Callable[[], V]]) -> None:
        self._values[key] = value

    def __delitem__(self, key: K) -> None:
        del self._values[key]

    def __iter__(self) -> Iterator[K]:
        return iter(self._values)

    def __len__(self) -> int:
        return len(self._values)


def relpath(path, start=os.curdir):
    path = os.fspath(path)
    start = os.path.abspath(os.fspath(start))

    # Windows path on different drive than curdir doesn't have relpath
    if os.name == "nt":
        # Since python 3.8 os.realpath resolves network shares to their UNC
        # path. So, to be certain that relative paths correctly captured,
        # we need to resolve to UNC path first. We resolve only the drive
        # name so that we don't follow any 'real' symlinks on the path
        def resolve_network_drive_windows(path_to_resolve):
            drive, tail = os.path.splitdrive(path_to_resolve)
            return os.path.join(os.path.realpath(drive), tail)

        path = resolve_network_drive_windows(os.path.abspath(path))
        start = resolve_network_drive_windows(start)
        if not os.path.commonprefix([start, path]):
            return path
    return os.path.relpath(path, start)



import os
from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping
from enum import Enum
from typing import TYPE_CHECKING, Callable, Optional, Union

from scmrepo.exceptions import SCMError
from scmrepo.git.objects import GitObject

if TYPE_CHECKING:
    from scmrepo.git.config import Config
    from scmrepo.git.objects import GitCommit, GitTag
    from scmrepo.progress import GitProgressEvent


class NoGitBackendError(SCMError):
    def __init__(self, func):
        super().__init__(f"No valid Git backend for '{func}'")


class SyncStatus(Enum):
    SUCCESS = 0
    UP_TO_DATE = 1
    DIVERGED = 2


class BaseGitBackend(ABC):
    """Base Git backend class."""

    @abstractmethod
    def __init__(self, root_dir=os.curdir, search_parent_directories=True):
        pass

    def close(self):  # noqa: B027
        pass

    @abstractmethod
    def is_ignored(self, path: str) -> bool:
        """Return True if the specified path is gitignored."""

    @property
    @abstractmethod
    def root_dir(self) -> Optional[str]:
        pass

    @staticmethod
    @abstractmethod
    def clone(
        url: str,
        to_path: Union[str, os.PathLike[str]],
        shallow_branch: Optional[str] = None,
        progress: Optional[Callable[["GitProgressEvent"], None]] = None,
        bare: bool = False,
        mirror: bool = False,
    ):
        pass

    @staticmethod
    @abstractmethod
    def init(path: str, bare: bool = False) -> None:
        pass

    @property
    @abstractmethod
    def dir(self) -> str:
        pass

    @abstractmethod
    def add(
        self,
        paths: Union[str, Iterable[str]],
        update: bool = False,
        force: bool = False,
    ):
        pass

    @abstractmethod
    def commit(self, msg: str, no_verify: bool = False):
        pass

    @abstractmethod
    def checkout(
        self,
        branch: str,
        create_new: Optional[bool] = False,
        force: bool = False,
        **kwargs,
    ):
        pass

    @abstractmethod
    def fetch(
        self,
        remote: Optional[str] = None,
        force: bool = False,
        unshallow: bool = False,
    ):
        pass

    @abstractmethod
    def pull(self, **kwargs):
        pass

    @abstractmethod
    def push(self):
        pass

    @abstractmethod
    def branch(self, branch: str):
        pass

    @abstractmethod
    def tag(
        self,
        tag: str,
        target: Optional[str] = None,
        annotated: bool = False,
        message: Optional[str] = None,
    ):
        pass

    @abstractmethod
    def untracked_files(self) -> Iterable[str]:
        pass

    @abstractmethod
    def is_tracked(self, path: str) -> bool:
        pass

    @abstractmethod
    def is_dirty(self, untracked_files: bool = False) -> bool:
        pass

    @abstractmethod
    def active_branch(self) -> str:
        pass

    @abstractmethod
    def active_branch_remote(self) -> str:
        """Return the fetch remote name for the current branch."""

    @abstractmethod
    def list_branches(self) -> Iterable[str]:
        pass

    @abstractmethod
    def list_tags(self) -> Iterable[str]:
        pass

    @abstractmethod
    def list_all_commits(self) -> Iterable[str]:
        pass

    @abstractmethod
    def get_tree_obj(self, rev: str, **kwargs) -> GitObject:
        pass

    @abstractmethod
    def get_rev(self) -> str:
        pass

    @abstractmethod
    def resolve_rev(self, rev: str) -> str:
        pass

    @abstractmethod
    def resolve_commit(self, rev: str) -> "GitCommit":
        pass

    @abstractmethod
    def set_ref(
        self,
        name: str,
        new_ref: str,
        old_ref: Optional[str] = None,
        message: Optional[str] = None,
        symbolic: Optional[bool] = False,
    ):
        """Set the specified git ref.

        Optional kwargs:
            old_ref: If specified, ref will only be set if it currently equals
                old_ref. Has no effect is symbolic is True.
            message: Optional reflog message.
            symbolic: If True, ref will be set as a symbolic ref to new_ref
                rather than the dereferenced object.
        """

    @abstractmethod
    def get_ref(self, name: str, follow: bool = True) -> Optional[str]:
        """Return the value of specified ref.

        If follow is false, symbolic refs will not be dereferenced.
        Returns None if the ref does not exist.
        """

    @abstractmethod
    def remove_ref(self, name: str, old_ref: Optional[str] = None):
        """Remove the specified git ref.

        If old_ref is specified, ref will only be removed if it currently
        equals old_ref.
        """

    @abstractmethod
    def iter_refs(self, base: Optional[str] = None):
        """Iterate over all refs in this git repo.

        If base is specified, only refs which begin with base will be yielded.
        """

    @abstractmethod
    def iter_remote_refs(self, url: str, base: Optional[str] = None, **kwargs):
        """Iterate over all refs in the specified remote Git repo.

        If base is specified, only refs which begin with base will be yielded.
        URL can be a named Git remote or URL.
        """

    @abstractmethod
    def get_refs_containing(self, rev: str, pattern: Optional[str] = None):
        """Iterate over all git refs containing the specified revision."""

    @abstractmethod
    def push_refspecs(
        self,
        url: str,
        refspecs: Union[str, Iterable[str]],
        force: bool = False,
        on_diverged: Optional[Callable[[str, str], bool]] = None,
        progress: Optional[Callable[["GitProgressEvent"], None]] = None,
        **kwargs,
    ) -> Mapping[str, SyncStatus]:
        """Push refspec to a remote Git repo.

        Args:
            url: Git remote name or absolute Git URL.
            refspecs: Iterable containing refspecs to push.
                Note that this will not match subkeys.
            force: If True, remote refs will be overwritten.
            on_diverged: Callback function which will be called if local ref
                and remote have diverged and force is False. If the callback
                returns True the remote ref will be overwritten.
                Callback will be of the form:
                    on_diverged(local_refname, remote_sha)
        """

    @abstractmethod
    def fetch_refspecs(
        self,
        url: str,
        refspecs: Union[str, Iterable[str]],
        force: bool = False,
        on_diverged: Optional[Callable[[str, str], bool]] = None,
        progress: Optional[Callable[["GitProgressEvent"], None]] = None,
        **kwargs,
    ) -> Mapping[str, SyncStatus]:
        """Fetch refspecs from a remote Git repo.

        Args:
            url: Git remote name or absolute Git URL.
            refspecs: Iterable containing refspecs to fetch.
                Note that this will not match subkeys.
            force: If True, local refs will be overwritten.
            on_diverged: Callback function which will be called if local ref
                and remote have diverged and force is False. If the callback
                returns True the local ref will be overwritten.
                Callback will be of the form:
                    on_diverged(local_refname, remote_sha)

        Returns:
            Mapping of local_refname to sync status.
        """

    @abstractmethod
    def _stash_iter(self, ref: str):
        """Iterate over stash commits in the specified ref."""

    @abstractmethod
    def _stash_push(
        self,
        ref: str,
        message: Optional[str] = None,
        include_untracked: bool = False,
    ) -> tuple[Optional[str], bool]:
        """Push a commit onto the specified stash.

        Returns a tuple of the form (rev, need_reset) where need_reset
        indicates whether or not the workspace should be `reset --hard`
        (some backends will not clean the workspace after creating a stash
        commit).
        """

    @abstractmethod
    def _stash_apply(
        self,
        rev: str,
        reinstate_index: bool = False,
        skip_conflicts: bool = False,
        **kwargs,
    ):
        """Apply the specified stash revision."""

    @abstractmethod
    def _stash_drop(self, ref: str, index: int):
        """Drop the specified stash revision."""

    @abstractmethod
    def _describe(
        self,
        revs: Iterable[str],
        base: Optional[str] = None,
        match: Optional[str] = None,
        exclude: Optional[str] = None,
    ) -> Mapping[str, Optional[str]]:
        """Return the first ref which points to each revs.

        Roughly equivalent to `git describe --all --exact-match`.

        If no matching ref was found, returns None.

        Optional kwargs:
            base: Base ref prefix to search, defaults to "refs/tags"
            match: Glob pattern. If specified only refs matching this glob
                pattern will be returned.
            exclude: Glob pattern. If specified, only refs which do not match
                this pattern will be returned.
        """

    @abstractmethod
    def diff(self, rev_a: str, rev_b: str, binary=False) -> str:
        """Return the git diff for two commits."""

    @abstractmethod
    def reset(self, hard: bool = False, paths: Optional[Iterable[str]] = None):
        """Reset current git HEAD."""

    @abstractmethod
    def checkout_index(
        self,
        paths: Optional[Iterable[str]] = None,
        force: bool = False,
        ours: bool = False,
        theirs: bool = False,
    ):
        """Checkout the specified paths from index."""

    @abstractmethod
    def status(
        self, ignored: bool = False, untracked_files: str = "all"
    ) -> tuple[Mapping[str, Iterable[str]], Iterable[str], Iterable[str]]:
        """Return tuple of (staged_files, unstaged_files, untracked_files).

        staged_files will be a dict mapping status (add, delete, modify) to a
        list of paths.

        If ignored is True, gitignored files will be included in
        untracked_paths.

        untracked_files can be one of the following:
          - "no" return no untracked files
          - "normal" (git cli default) return untracked files and directories
          - "all" (default) return all untracked files in untracked directories

        Using "no" or "normal" will be faster than "all" when large untracked
        directories are present in the workspace, as collecting all untracked
        files can take some time.
        """

    def _reset(self) -> None:  # noqa: B027
        pass

    @abstractmethod
    def merge(
        self,
        rev: str,
        commit: bool = True,
        msg: Optional[str] = None,
        squash: bool = False,
    ) -> Optional[str]:
        """Merge specified commit into the current (working copy) branch.

        Args:
            rev: Git revision to merge.
            commit: If True, merge will be committed using `msg` as the commit
                message. If False, the merge changes will be staged but not
                committed.
            msg: Merge commit message, required if `commit` is True.
            squash: If True, working tree state will be updated with merge
                results, but no commit will be made and no changes will be
                staged. `commit` must be False when `squash` is True.

        If conflicts are present after merging, MergeConflictError will be
        raised. The caller is responsible for either resolving the conflicts
        or resetting the repository state.

        Returns revision of the merge commit or None if no commit was made.
        """

    @abstractmethod
    def validate_git_remote(self, url: str, **kwargs):
        """Verify that url is a valid git URL or remote name."""

    @abstractmethod
    def get_remote_url(self, remote: str) -> str:
        """Return URL for the specified remote."""

    @abstractmethod
    def check_ref_format(self, refname: str) -> bool:
        """Check if a reference name is well formed."""

    @abstractmethod
    def get_tag(self, name: str) -> Optional[Union[str, "GitTag"]]:
        """Return the specified tag object.

        Args:
            name: Tag name (without 'refs/tags/' prefix).

        Returns:
            None if the specified tag does not exist.
            String SHA for the target object if the tag is a lightweight tag.
            GitTag object if the tag is an annotated tag.
        """

    @abstractmethod
    def get_config(self, path: Optional[str] = None) -> "Config":
        """Return a Git config object.

        Args:
            path: If set, a config object for the specified config file will be
                returned. By default, the standard Git system/global/repo config
                stack object will be returned.
        """

    @abstractmethod
    def check_attr(
        self,
        path: str,
        attr: str,
        source: Optional[str] = None,
    ) -> Optional[Union[bool, str]]:
        """Return the value of the specified attribute for a pathname.

        Args:
            path: Pathname to check.
            attr: Attribute to check.
            source: Optional tree-ish source to check.

        Returns:
            None when the attribute is not defined for the path (unspecified).
            True when the attribute is defined as true (set).
            False when the attribute is defined as false (unset).
            The value of the attribute when a value has been assigned.
        """



from collections.abc import Iterator
from typing import Optional, Union

from dulwich.client import HTTPUnauthorized, Urllib3HttpGitClient

from scmrepo.git.credentials import Credential, CredentialNotFoundError


class GitCredentialsHTTPClient(Urllib3HttpGitClient):  # pylint: disable=abstract-method
    def __init__(
        self,
        base_url,
        username=None,
        password=None,
        config=None,
        **kwargs,
    ) -> None:
        super().__init__(
            base_url=base_url,
            username=username,
            password=password,
            config=config,
            **kwargs,
        )
        self._store_credentials: Optional[Credential] = None

    def _http_request(
        self,
        url: str,
        headers: Optional[dict[str, str]] = None,
        data: Optional[Union[bytes, Iterator[bytes]]] = None,
        raise_for_status: bool = True,
    ):
        cached_chunks: list[bytes] = []

        def _cached_data() -> Iterator[bytes]:
            assert data is not None
            if isinstance(data, bytes):
                yield data
                return

            if cached_chunks:
                yield from cached_chunks
                return

            for chunk in data:
                cached_chunks.append(chunk)
                yield chunk

        try:
            result = super()._http_request(
                url,
                headers=headers,
                data=None if data is None else _cached_data(),
                raise_for_status=raise_for_status,
            )
        except HTTPUnauthorized:
            auth_header = self._get_auth()
            if not auth_header:
                raise
            if headers:
                headers.update(auth_header)
            else:
                headers = auth_header
            result = super()._http_request(
                url,
                headers=headers,
                data=None if data is None else _cached_data(),
                raise_for_status=raise_for_status,
            )
        if self._store_credentials is not None:
            self._store_credentials.approve()
        return result

    def _get_auth(self) -> dict[str, str]:
        from urllib3.util import make_headers

        try:
            base_url = self._base_url.rstrip("/")
            creds = Credential(username=self._username, url=base_url).fill()
            self._store_credentials = creds
            return make_headers(basic_auth=f"{creds.username}:{creds.password}")
        except CredentialNotFoundError:
            return {}



from collections.abc import Callable

import pytest
from pytest_test_utils import TmpDir
from pytest_test_utils.matchers import Matcher

from scmrepo.git import Git
from scmrepo.noscm import NoSCM


def test_noscm(tmp_dir: TmpDir):
    scm = NoSCM(tmp_dir)
    scm.add("test")


def test_noscm_raises_exc_on_unimplemented_apis(tmp_dir: TmpDir, matcher: Matcher):
    class Unimplemented(Exception):
        pass

    scm = NoSCM(tmp_dir, _raise_not_implemented_as=Unimplemented)
    assert scm._exc is Unimplemented  # pylint: disable=protected-access

    assert Git.reset == matcher.instance_of(Callable)
    with pytest.raises(Unimplemented):
        assert scm.reset()



from .client import LFSClient
from .exceptions import LFSError
from .fetch import fetch
from .pointer import Pointer
from .smudge import smudge
from .storage import LFSStorage

__all__ = ["LFSClient", "LFSError", "LFSStorage", "Pointer", "fetch", "smudge"]



# credentials.py -- support for git credential helpers

# Copyright (C) 2022 Daniele Trifirò <daniele@iterative.ai>
#
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as public by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Support for git credential helpers

https://git-scm.com/book/en/v2/Git-Tools-Credential-Storage

Currently Dulwich supports only the `get` operation

"""

import locale
import logging
import os
import shlex
import shutil
import subprocess  # nosec B404
import sys
from abc import ABC, abstractmethod
from collections.abc import Iterator, Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    NamedTuple,
    Optional,
    Union,
)
from urllib.parse import urlparse, urlunparse

from dulwich.config import StackedConfig
from dulwich.credentials import urlmatch_credential_sections
from funcy import cached_property

from scmrepo.exceptions import SCMError

if TYPE_CHECKING:
    from collections.abc import Iterable

    from dulwich.config import ConfigDict

logger = logging.getLogger(__name__)

SectionLike = Union[bytes, str, tuple[Union[bytes, str], ...]]


class CredentialNotFoundError(SCMError):
    """Error occurred while retrieving credentials/no credentials available."""


class CredentialQuitError(SCMError):
    """Credential helper returned quit=1."""


class CredentialHelper(ABC):
    """Base git-credential helper."""

    @abstractmethod
    def get(self, credential: "Credential", **kwargs) -> "Credential":
        """Get a matching credential from this helper.

        Raises:
            CredentialNotFoundError: No matching credential was found.
        """

    @abstractmethod
    def store(self, credential: "Credential", **kwargs):
        """Store the credential, if applicable to the helper"""

    @abstractmethod
    def erase(self, credential: "Credential", **kwargs):
        """Remove a matching credential, if any, from the helper's storage"""


class GitCredentialHelper(CredentialHelper):
    """Helper for retrieving credentials through git-credential-<helper> commands.

    Users should generally not need to use credential helpers directly.
    The corresponding Credential methods should be used instead.
    (i.e. Credential.fill() instead of GitCredentialHelper.get())

    Usage:
    >>> helper = GitCredentialHelper("store") # Use `git credential-store`
    >>> generated = Credential(url="https://github.com/dtrifiro/aprivaterepo")
    >>> credentials = helper.get(generated)
    >>> username = credentials.username
    >>> password = credentials.password
    """

    def __init__(self, command: str, use_http_path: bool = False) -> None:
        super().__init__()
        self._command = command
        self._run_kwargs: dict[str, Any] = {}
        if self._command[0] == "!":
            # On Windows this will only work in git-bash and/or WSL2
            self._run_kwargs["shell"] = True
        self._encoding = locale.getpreferredencoding()
        self.use_http_path = use_http_path

    def _prepare_command(self, action: Optional[str] = None) -> Union[str, list[str]]:
        if self._command[0] == "!":
            return self._command[1:] + (f" {action}" if action else "")

        if sys.platform != "win32":
            argv = shlex.split(self._command)
        else:
            # On windows, subprocess.run uses subprocess.list2cmdline() to
            # join arguments when providing a list, so we can just split
            # using whitespace.
            argv = self._command.split()
        if action:
            argv.append(action)

        if os.path.isabs(argv[0]):
            return argv

        executable = f"git-credential-{argv[0]}"
        if not shutil.which(executable) and shutil.which("git"):
            # If the helper cannot be found in PATH, it might be
            # a C git helper in GIT_EXEC_PATH
            git_exec_path = subprocess.check_output(
                ("git", "--exec-path"),
                text=True,
            ).strip()
            if shutil.which(executable, path=git_exec_path):
                executable = os.path.join(git_exec_path, executable)

        return [executable, *argv[1:]]

    def get(self, credential: "Credential", **kwargs) -> "Credential":
        if not (credential.protocol or credential.host):
            raise ValueError("One of protocol, hostname must be provided")
        cmd = self._prepare_command("get")
        use_path = credential.protocol in ("http", "https") and self.use_http_path
        helper_input = [
            f"{key}={value}"
            for key, value in credential.items()
            if key != "path" or use_path
        ]
        helper_input.append("")

        try:
            res = subprocess.run(  # noqa: S603
                cmd,
                check=True,
                capture_output=True,
                input="\n".join(helper_input),
                encoding=self._encoding,
                **self._run_kwargs,
            )
        except subprocess.CalledProcessError as exc:
            raise CredentialNotFoundError(exc.stderr) from exc
        except FileNotFoundError as exc:
            raise CredentialNotFoundError("Helper not found") from exc
        if res.stderr:
            logger.debug(res.stderr)

        credentials: dict[str, Any] = {}
        for line in res.stdout.splitlines():
            try:
                key, value = line.split("=", maxsplit=1)
                # Only include credential values that are used in the Credential
                # constructor.
                # Other values may be returned by the subprocess, but they must be
                # ignored.
                # e.g. osxkeychain credential helper >= 2.46.0 can return
                # `capability[]` and `state`)
                if key in [
                    "protocol",
                    "host",
                    "path",
                    "username",
                    "password",
                    "password_expiry_utc",
                    "url",
                ]:
                    # Garbage bytes were output from git-credential-osxkeychain from
                    # 2.45.0 to 2.47.0:
                    # https://github.com/git/git/commit/6c3c451fb6e1c3ca83f74e63079d4d0af01b2d69
                    credentials[key] = _strip_garbage_bytes(value)
            except ValueError:
                continue
        if not credentials:
            raise CredentialNotFoundError("No credentials found")
        quit_ = credentials.get("quit")
        if quit_ is not None and quit_.lower() in ("true", "1"):
            raise CredentialQuitError("Helper returned quit=1")
        return Credential(**credentials)

    def store(self, credential: "Credential", **kwargs):
        """Store the credential, if applicable to the helper"""
        cmd = self._prepare_command("store")
        use_path = credential.protocol in ("http", "https") and self.use_http_path
        helper_input = [
            f"{key}={value}"
            for key, value in credential.items()
            if key != "path" or use_path
        ]
        helper_input.append("")

        try:
            res = subprocess.run(  # noqa: S603
                cmd,
                capture_output=True,
                input="\n".join(helper_input),
                encoding=self._encoding,
                **self._run_kwargs,
                check=False,
            )
            if res.stderr:
                logger.debug(res.stderr)
        except FileNotFoundError:
            logger.debug("Helper not found", exc_info=True)

    def erase(self, credential: "Credential", **kwargs):
        """Remove a matching credential, if any, from the helper's storage"""
        cmd = self._prepare_command("erase")
        use_path = credential.protocol in ("http", "https") and self.use_http_path
        helper_input = [
            f"{key}={value}"
            for key, value in credential.items()
            if key != "path" or use_path
        ]
        helper_input.append("")

        try:
            res = subprocess.run(  # noqa: S603
                cmd,
                capture_output=True,
                input="\n".join(helper_input),
                encoding=self._encoding,
                **self._run_kwargs,
                check=False,
            )
            if res.stderr:
                logger.debug(res.stderr)
        except FileNotFoundError:
            logger.debug("Helper not found", exc_info=True)

    @staticmethod
    def get_matching_commands(
        base_url: str, config: Optional[Union["ConfigDict", "StackedConfig"]] = None
    ) -> Iterator[tuple[str, bool]]:
        config = config or StackedConfig.default()
        if isinstance(config, StackedConfig):
            backends: Iterable[ConfigDict] = config.backends
        else:
            backends = [config]

        for conf in backends:
            # We will try to match credential sections' url with the given url,
            # falling back to the generic section if there's no match
            for section in urlmatch_credential_sections(conf, base_url):
                try:
                    command = conf.get(section, "helper")
                except KeyError:
                    # no helper configured
                    continue
                if not command:
                    continue
                use_http_path = conf.get_boolean(section, "usehttppath", False)
                yield (
                    command.decode(conf.encoding or sys.getdefaultencoding()),
                    use_http_path,
                )


def _strip_garbage_bytes(s: str) -> str:
    """
    Garbage (random) bytes were output from git-credential-osxkeychain from
    2.45.0 to 2.47.0 so must be removed.
    https://github.com/git/git/commit/6c3c451fb6e1c3ca83f74e63079d4d0af01b2d69
    :param s: string that might contain garbage/random bytes
    :return str: The string with the garbage bytes removed
    """
    # Assume that any garbage bytes begin with a 0-byte
    zero = s.find(chr(0))
    return s[0:zero] if zero >= 0 else s


class _CredentialKey(NamedTuple):
    protocol: str
    host: Optional[str]
    path: Optional[str]

    @classmethod
    def from_credential(cls, credential: "Credential"):
        return cls(credential.protocol or "", credential.host, credential.path)


def _input_tty(prompt: str = "Username: ") -> str:
    """Prompt for username on /dev/tty when available.

    Defaults to builtin input() when /dev/tty cannot be opened, does not disable
    echo (use getpass.getpass() instead).
    """
    import io
    from contextlib import ExitStack

    if os.name == "nt":
        if not sys.stdin.isatty():
            raise EOFError
        return input(prompt)

    with ExitStack() as stack:
        try:
            fd = os.open(
                "/dev/tty",
                os.O_RDWR | os.O_NOCTTY,  # type: ignore[attr-defined]
            )
            tty = io.FileIO(fd, "w+")
            stack.enter_context(tty)
            stream = io.TextIOWrapper(tty)
            stack.enter_context(stream)
        except OSError:
            stack.close()
            # fallback to default input()
            if not sys.stdin.isatty():
                raise
            return input(prompt)
        try:
            stream.write(prompt)
            stream.flush()
            line = stream.readline()
            if not line:
                raise EOFError
            if line[-1] == "\n":
                line = line[:-1]
            return line
        finally:
            stream.flush()
    raise EOFError


class MemoryCredentialHelper(CredentialHelper):
    """Memory credential helper that supports optional interactive input."""

    def __init__(self) -> None:
        super().__init__()
        self._credentials: dict[_CredentialKey, Credential] = {}

    def __getitem__(self, key: object) -> "Credential":
        if isinstance(key, _CredentialKey):
            return self._credentials[key]
        if isinstance(key, Credential):
            return self._credentials[_CredentialKey.from_credential(key)]
        raise KeyError

    def __setitem__(self, key: object, value: "Credential"):
        if isinstance(key, _CredentialKey):
            self._credentials[key] = value
        elif isinstance(key, Credential):
            self._credentials[_CredentialKey.from_credential(key)] = value
        else:
            raise ValueError

    def __delitem__(self, key: object):
        if isinstance(key, _CredentialKey):
            del self._credentials[key]
        elif isinstance(key, Credential):
            del self._credentials[_CredentialKey.from_credential(key)]
        else:
            raise ValueError

    def get(
        self, credential: "Credential", *, interactive: bool = False, **kwargs
    ) -> "Credential":
        """Get a matching credential from this helper.

        Raises:
            CredentialNotFoundError: No matching credential was found.
        """
        from getpass import getpass

        if credential.path:
            try_creds = [
                credential,
                Credential(protocol=credential.protocol, host=credential.host),
            ]
        else:
            try_creds = [credential]
        for cred in try_creds:
            try:
                return self[cred]
            except KeyError:
                pass
        if interactive:
            try:
                if self.askpass:
                    return self._get_interactive(credential, self.askpass.input)
                if os.environ.get("GIT_TERMINAL_PROMPT") != "0":
                    return self._get_interactive(credential, _input_tty, getpass)
            except (EOFError, OSError):
                pass
        raise CredentialNotFoundError("No matching credentials")

    def _get_interactive(
        self,
        credential: "Credential",
        input_echo: Callable[[str], str],
        input_noecho: Optional[Callable[[str], str]] = None,
    ) -> "Credential":
        if not input_noecho:
            input_noecho = input_echo
        new = Credential(
            protocol=credential.protocol,
            host=credential.host,
            path=credential.path,
            username=credential.username,
            password=credential.password,
        )
        try:
            if not new.username:
                prompt = f"Username for '{new.describe()}': "
                new.username = input_echo(prompt)
            if not new.password:
                prompt = f"Password for '{new.describe()}': "
                new.password = input_noecho(prompt)
        except KeyboardInterrupt as exc:
            raise CredentialNotFoundError("User cancelled prompt") from exc
        return new

    def store(self, credential: "Credential", **kwargs):
        """Store the credential, if applicable to the helper"""
        if credential.protocol or credential.host or credential.path:
            self[credential] = credential

    def erase(self, credential: "Credential", **kwargs):
        """Remove a matching credential, if any, from the helper's storage"""
        try:
            del self[credential]
        except KeyError:
            pass

    @cached_property
    def askpass(self) -> Optional["_AskpassCommand"]:
        return self.get_askpass()

    @staticmethod
    def get_askpass(
        config: Optional[Union["ConfigDict", "StackedConfig"]] = None,
    ) -> Optional["_AskpassCommand"]:
        askpass = os.environ.get("GIT_ASKPASS")
        if not askpass:
            config = config or StackedConfig.default()
            try:
                askpass = config.get("core", "askpass").decode(sys.getdefaultencoding())
            except KeyError:
                pass
        if not askpass:
            askpass = os.environ.get("SSH_ASKPASS")
        if askpass:
            return _AskpassCommand(askpass)
        return None


class _AskpassCommand:
    def __init__(self, command: str):
        self.command = command

    def input(self, prompt: str) -> str:
        argv = [self.command, prompt]
        try:
            res = subprocess.run(  # noqa: S603
                argv,
                check=True,
                capture_output=True,
                encoding=locale.getpreferredencoding(),
            )
        except (FileNotFoundError, subprocess.CalledProcessError):
            return ""
        return res.stdout.splitlines()[0]


memory_helper = MemoryCredentialHelper()


class Credential(Mapping[str, str]):
    """Git credentials, equivalent to CGit git-credential API.

    Usage:

    1. Generate a credential based on context

        >>> generated = Credential(url="https://github.com/dtrifiro/aprivaterepo")

    2. Ask git-credential to give username/password for this context

        >>> credential = generated.fill()

    3. Use the credential from (2) in Git operation
    4. If the operation in (3) was successful, approve it for reuse in subsequent
       operations

       >>> credential.approve()

    See also:
        https://git-scm.com/docs/git-credential#_typical_use_of_git_credential
        https://github.com/git/git/blob/master/credential.h

    """

    _SUPPORTED_KEYS = (
        "protocol",
        "host",
        "path",
        "username",
        "password",
        "password_expiry_utc",
    )

    def __init__(
        self,
        *,
        protocol: Optional[str] = None,
        host: Optional[str] = None,  # host with optional ":<port>" included
        path: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        password_expiry_utc: Optional[str] = None,
        url: Optional[str] = None,
    ):
        self.protocol = protocol
        self.host = host
        self.path = path
        self.username = username
        self.password = password
        self.password_expiry_utc = password_expiry_utc
        self._approved = False
        if url:
            parsed = urlparse(url)
            self.protocol = self.protocol or parsed.scheme
            if not self.protocol:
                raise ValueError("protocol must be specified when using URL")
            port = f":{parsed.port}" if parsed.port is not None else ""
            hostname = parsed.hostname or ""
            self.host = self.host or f"{hostname}{port}"
            self.username = self.username or parsed.username
            self.password = self.password or parsed.password
            if parsed.path:
                self.path = self.path or parsed.path.lstrip("/")

    def __getitem__(self, key: object) -> str:
        if isinstance(key, str):
            try:
                return getattr(self, key)
            except AttributeError:
                pass
        raise KeyError

    def __iter__(self) -> Iterator[str]:
        for key in self._SUPPORTED_KEYS:
            try:
                value = self[key]
                if value is not None:
                    yield key
            except KeyError:
                pass

    def __len__(self) -> int:
        return len(list(iter(self)))

    def __str__(self) -> str:
        return self.describe(use_path=True, sanitize=True)

    @property
    def url(self) -> str:
        return self.describe(use_path=True, sanitize=False)

    def describe(self, use_path: bool = False, sanitize: bool = True) -> str:
        username = self.username or ""
        if sanitize:
            password = ":***" if self.password else ""
        else:
            password = f":{self.password}" if self.password else ""
        at = "@" if (username or password) else ""
        host = f"{self.host}" if self.host else ""
        netloc = f"{username}{password}{at}{host}"
        if use_path:
            path = self.path or ""
        else:
            path = ""
        return urlunparse((self.protocol or "", netloc, path, "", "", ""))

    @cached_property
    def helpers(self) -> list["CredentialHelper"]:
        url = self.url
        return [
            GitCredentialHelper(command, use_http_path=use_http_path)
            for command, use_http_path in GitCredentialHelper.get_matching_commands(url)
        ]

    def fill(self, interactive: bool = True) -> "Credential":
        """Return a new credential with filled username and password."""
        if self.username and self.password:
            return Credential(**self)

        try:
            return memory_helper.get(self, interactive=False)
        except CredentialNotFoundError:
            pass

        for helper in self.helpers:
            try:
                return helper.get(self)
            except CredentialNotFoundError:
                continue
            except CredentialQuitError as exc:
                raise CredentialNotFoundError(
                    f"No available credentials for '{self}'"
                ) from exc

        try:
            return memory_helper.get(self, interactive=interactive)
        except CredentialNotFoundError:
            pass

        raise CredentialNotFoundError(f"No available credentials for '{self}'")

    def approve(self):
        """Store this credential in available helpers."""
        if self._approved or not (self.username and self.password):
            return
        for helper in self.helpers:
            helper.store(self)
        memory_helper.store(self)
        self._approved = True

    def reject(self):
        """Remove this credential from available helpers."""
        for helper in self.helpers:
            helper.erase(self)
        memory_helper.erase(self)
        self._approved = False



import hashlib
import io
import logging
from dataclasses import dataclass
from typing import IO, BinaryIO, TextIO

logger = logging.getLogger(__name__)


LFS_VERSION = "https://git-lfs.github.com/spec/v1"
LEGACY_LFS_VERSION = "https://hawser.github.com/spec/v1"
ALLOWED_VERSIONS = (LFS_VERSION, LEGACY_LFS_VERSION)
HEADERS = [(b"version " + version.encode("utf-8")) for version in ALLOWED_VERSIONS]


def _get_kv(line: str) -> tuple[str, str]:
    key, value = line.strip().split(maxsplit=1)
    return key, value


@dataclass
class Pointer:
    oid: str
    size: int

    def __init__(self, oid: str, size: int, **kwargs):
        self.oid = oid
        self.size = size
        self._dict = kwargs

    def __hash__(self):
        return hash(self.dump())

    @classmethod
    def build(cls, fobj: BinaryIO) -> "Pointer":
        m = hashlib.sha256()
        data = fobj.read()
        m.update(data)
        return cls(m.hexdigest(), len(data))

    @classmethod
    def load(cls, fobj: IO) -> "Pointer":
        """Load the specified pointer file."""

        if isinstance(fobj, io.TextIOBase):  # type: ignore[unreachable]
            text_obj: TextIO = fobj  # type: ignore[unreachable]

        else:
            text_obj = io.TextIOWrapper(fobj, encoding="utf-8")

        cls.check_version(text_obj.readline())
        d = dict(_get_kv(line) for line in text_obj.readlines())
        try:
            value = d.pop("oid")
            hash_method, oid = value.split(":", maxsplit=1)
            if hash_method != "sha256":
                raise ValueError("Invalid LFS hash method '{hash_method}'")
        except ValueError as e:
            raise ValueError("Invalid LFS pointer oid") from e
        try:
            value = d.pop("size")
            size = int(value)
        except ValueError as e:
            raise ValueError("Invalid LFS pointer size") from e

        return cls(oid, size, **d)

    @staticmethod
    def check_version(line: str):
        try:
            key, value = _get_kv(line)
            if key != "version":
                raise ValueError("LFS pointer file must start with 'version'")
            if value not in ALLOWED_VERSIONS:
                raise ValueError(f"Unsupported LFS pointer version '{value}'")
        except (ValueError, OSError) as e:
            raise ValueError("Invalid LFS pointer file") from e

    def dump(self) -> str:
        d = {
            "oid": f"sha256:{self.oid}",
            "size": self.size,
        }
        d.update(self._dict)
        return "\n".join(
            [f"version {LFS_VERSION}"] + [f"{key} {d[key]}" for key in sorted(d)] + [""]
        )

    def to_bytes(self) -> bytes:
        return self.dump().encode("utf-8")


if __name__ == "__main__":
    # Minimal `git lfs pointer` CLI implementation
    import argparse
    import sys

    parser = argparse.ArgumentParser(
        description="Build generated pointer files.",
    )
    parser.add_argument("--file", help="A local file to build the pointer from.")
    args = parser.parse_args()
    if not args.file:
        sys.exit("Nothing to do")

    with open(args.file, mode="rb") as fobj_:
        p = Pointer.build(fobj_)



"""Manages Git."""

import logging
import os
import re
import typing
from collections import OrderedDict
from collections.abc import Iterable, Mapping
from contextlib import contextmanager
from functools import partialmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Optional,
    Union,
)

from funcy import cached_property, first
from pathspec.patterns import GitWildMatchPattern

from scmrepo.base import Base
from scmrepo.exceptions import FileNotInRepoError, GitHookAlreadyExists, RevError
from scmrepo.utils import relpath

from .backend.base import BaseGitBackend, NoGitBackendError, SyncStatus
from .backend.dulwich import DulwichBackend
from .backend.gitpython import GitPythonBackend
from .backend.pygit2 import Pygit2Backend
from .objects import (  # noqa: F401, pylint: disable=unused-import
    GitCommit,
    GitObject,
    GitTag,
    GitTrie,
)
from .stash import Stash

if TYPE_CHECKING:
    from scmrepo.progress import GitProgressEvent

logger = logging.getLogger(__name__)

BackendCls = type[BaseGitBackend]


_LOW_PRIO_BACKENDS = ("gitpython",)


class GitBackends(Mapping):
    DEFAULT: ClassVar[dict[str, BackendCls]] = {
        "dulwich": DulwichBackend,
        "pygit2": Pygit2Backend,
        "gitpython": GitPythonBackend,
    }

    def __getitem__(self, key: str) -> BaseGitBackend:
        """Lazily initialize backends and cache it afterwards"""
        initialized = self.initialized.get(key)
        if not initialized:
            if key not in self.backends and key in self.DEFAULT:
                raise NotImplementedError
            backend = self.backends[key]
            initialized = backend(*self.args, **self.kwargs)
            self.initialized[key] = initialized
        return initialized

    def __init__(self, selected: Optional[Iterable[str]], *args, **kwargs) -> None:
        selected = selected or list(self.DEFAULT)
        self.backends = OrderedDict((key, self.DEFAULT[key]) for key in selected)

        self.initialized: dict[str, BaseGitBackend] = {}

        self.args = args
        self.kwargs = kwargs

    def __iter__(self):
        return iter(self.backends)

    def __len__(self) -> int:
        return len(self.backends)

    def close_initialized(self) -> None:
        for backend in self.initialized.values():
            backend.close()

    def reset_all(self) -> None:
        for backend in self.initialized.values():
            backend._reset()  # pylint: disable=protected-access

    def move_to_end(self, key: str, last: bool = True):
        if key not in _LOW_PRIO_BACKENDS:
            self.backends.move_to_end(key, last=last)


class Git(Base):
    """Class for managing Git."""

    GITIGNORE = ".gitignore"
    GIT_DIR = ".git"
    LOCAL_BRANCH_PREFIX = "refs/heads/"
    RE_HEXSHA = re.compile(r"^[0-9A-Fa-f]{4,40}$")
    BAD_REF_CHARS_RE = re.compile("[\177\\s~^:?*\\[]")

    def __init__(self, *args, backends: Optional[Iterable[str]] = None, **kwargs):
        self.backends = GitBackends(backends, *args, **kwargs)
        first_ = first(self.backends.values())
        super().__init__(first_.root_dir)
        self._last_backend: Optional[str] = None

    @property
    def dir(self):
        for backend in self.backends.values():
            try:
                return backend.dir
            except NotImplementedError:
                pass
        raise NotImplementedError

    @cached_property
    def hooks_dir(self):
        from pathlib import Path

        return Path(self.root_dir) / self.GIT_DIR / "hooks"

    @property
    def gitpython(self):
        return self.backends["gitpython"]

    @property
    def dulwich(self):
        return self.backends["dulwich"]

    @property
    def pygit2(self):
        return self.backends["pygit2"]

    @cached_property
    def stash(self):
        return Stash(self)

    @classmethod
    def clone(
        cls,
        url: str,
        to_path: Union[str, os.PathLike[str]],
        rev: Optional[str] = None,
        bare: bool = False,
        mirror: bool = False,
        **kwargs,
    ):
        assert not (rev and bare), "rev checkout is unsupported in bare repos"
        for _, backend in GitBackends.DEFAULT.items():
            try:
                backend.clone(url, to_path, bare=bare, mirror=mirror, **kwargs)
                repo = cls(to_path)
                if rev:
                    repo.checkout(rev)
                return repo
            except NotImplementedError:
                pass
        raise NoGitBackendError("clone")

    @classmethod
    def is_sha(cls, rev):
        return rev and cls.RE_HEXSHA.search(rev)

    @classmethod
    def split_ref_pattern(cls, ref: str) -> tuple[str, str]:
        name = cls.BAD_REF_CHARS_RE.split(ref, maxsplit=1)[0]
        return name, ref[len(name) :]

    @staticmethod
    def _get_git_dir(root_dir):
        return os.path.join(root_dir, Git.GIT_DIR)

    @property
    def ignore_file(self):
        return self.GITIGNORE

    @cached_property
    def lfs_storage(self):
        from .lfs import LFSStorage
        from .lfs.storage import get_storage_path

        return LFSStorage(get_storage_path(self))

    def _get_gitignore(self, path):
        ignore_file_dir = os.path.dirname(path)

        assert os.path.isabs(path)
        assert os.path.isabs(ignore_file_dir)

        entry = relpath(path, ignore_file_dir).replace(os.sep, "/")
        # NOTE: using '/' prefix to make path unambiguous
        if len(entry) > 0 and entry[0] != "/":
            entry = "/" + entry

        gitignore = os.path.join(ignore_file_dir, self.GITIGNORE)

        if not os.path.realpath(gitignore).startswith(self.root_dir + os.sep):
            raise FileNotInRepoError(
                f"'{path}' is outside of git repository '{self.root_dir}'"
            )

        return entry, gitignore

    def ignore(self, path: str) -> Optional[str]:
        entry, gitignore = self._get_gitignore(path)
        if self.is_ignored(path):
            return None

        self._add_entry_to_gitignore(entry, gitignore)
        return gitignore

    def _add_entry_to_gitignore(self, entry, gitignore):
        entry = GitWildMatchPattern.escape(entry)

        with open(gitignore, "a+", encoding="utf-8") as fobj:
            unique_lines = set(fobj.readlines())
            fobj.seek(0, os.SEEK_END)
            if fobj.tell() == 0:
                # Empty file
                prefix = ""
            else:
                fobj.seek(fobj.tell() - 1, os.SEEK_SET)
                last = fobj.read(1)
                prefix = "" if last == "\n" else "\n"
            new_entry = f"{prefix}{entry}\n"
            if new_entry not in unique_lines:
                fobj.write(new_entry)

    def ignore_remove(self, path: str) -> Optional[str]:
        entry, gitignore = self._get_gitignore(path)

        if not os.path.exists(gitignore):
            return None

        with open(gitignore, encoding="utf-8") as fobj:
            lines = fobj.readlines()

        filtered = list(filter(lambda x: x.strip() != entry.strip(), lines))

        if not filtered:
            os.unlink(gitignore)
            return None

        with open(gitignore, "w", encoding="utf-8") as fobj:
            fobj.writelines(filtered)
        return gitignore

    def verify_hook(self, name):
        if (self.hooks_dir / name).exists():
            raise GitHookAlreadyExists(name)

    def install_hook(self, name: str, script: str, interpreter: str = "sh") -> None:
        import shutil

        self.hooks_dir.mkdir(exist_ok=True)
        hook = self.hooks_dir / name

        directive = f"#!{shutil.which(interpreter) or '/bin/sh'}"
        hook.write_text(f"{directive}\n{script}\n", encoding="utf-8")
        hook.chmod(0o777)

    def install_merge_driver(self, name: str, description: str, driver: str) -> None:
        self.gitpython.repo.git.config(f"merge.{name}.name", description)
        self.gitpython.repo.git.config(f"merge.{name}.driver", driver)

    def belongs_to_scm(self, path):
        basename = os.path.basename(path)
        path_parts = os.path.normpath(path).split(os.path.sep)
        return basename == self.ignore_file or Git.GIT_DIR in path_parts

    def has_rev(self, rev):
        try:
            self.resolve_rev(rev)
            return True
        except RevError:
            return False

    def close(self):
        self.backends.close_initialized()
        if "lfs_storage" in self.__dict__:
            self.lfs_storage.close()

    @property
    def no_commits(self):
        return not bool(self.get_ref("HEAD"))

    # Prefer reusing the most recently used backend when possible. When
    # changing backends (due to unimplemented calls), we close the previous
    # backend to release any open git files/contexts that may cause conflicts
    # with the new backend.
    #
    # See:
    # https://github.com/iterative/dvc/issues/5641
    # https://github.com/iterative/dvc/issues/7458
    def _backend_func(self, name, *args, **kwargs) -> Any:
        backends: Iterable[str] = kwargs.pop("backends", self.backends)
        for key in backends:
            if self._last_backend is not None and key != self._last_backend:
                self.backends[self._last_backend].close()
                self._last_backend = None
            try:
                backend = self.backends[key]
                func = getattr(backend, name)
                result = func(*args, **kwargs)
                self._last_backend = key
                self.backends.move_to_end(key, last=False)
                return result
            except NotImplementedError:
                pass
        raise NoGitBackendError(name)

    def get_fs(self, rev: str):
        from scmrepo.fs import GitFileSystem

        return GitFileSystem(scm=self, rev=rev)

    @classmethod
    def init(
        cls,
        path: Union[str, os.PathLike[str]],
        bare: bool = False,
        _backend: Optional[str] = None,
    ) -> "Git":
        for name, backend in GitBackends.DEFAULT.items():
            if _backend and name != _backend:
                continue
            try:
                backend.init(os.fsdecode(path), bare=bare)
                # TODO: reuse created object instead of initializing a new one.
                return cls(path)
            except NotImplementedError:
                pass
        raise NoGitBackendError("init")

    def add_commit(
        self,
        paths: Union[str, Iterable[str]],
        message: str,
    ) -> None:
        self.add(paths)
        self.commit(msg=message)

    _fetch_refspecs = partialmethod(
        _backend_func, "fetch_refspecs", backends=["pygit2", "dulwich"]
    )

    def fetch_refspecs(
        self,
        url: str,
        refspecs: Union[str, Iterable[str]],
        force: bool = False,
        on_diverged: Optional[Callable[[str, str], bool]] = None,
        progress: Optional[Callable[["GitProgressEvent"], None]] = None,
        **kwargs,
    ) -> typing.Mapping[str, SyncStatus]:
        return self._fetch_refspecs(
            url,
            refspecs,
            force=force,
            on_diverged=on_diverged,
            progress=progress,
            **kwargs,
        )

    is_ignored = partialmethod(_backend_func, "is_ignored")
    add = partialmethod(_backend_func, "add")
    commit = partialmethod(_backend_func, "commit")
    checkout = partialmethod(_backend_func, "checkout")
    fetch = partialmethod(_backend_func, "fetch")
    pull = partialmethod(_backend_func, "pull")
    push = partialmethod(_backend_func, "push")
    branch = partialmethod(_backend_func, "branch")
    tag = partialmethod(_backend_func, "tag")
    untracked_files = partialmethod(_backend_func, "untracked_files")
    is_tracked = partialmethod(_backend_func, "is_tracked")
    is_dirty = partialmethod(_backend_func, "is_dirty")
    active_branch = partialmethod(_backend_func, "active_branch")
    active_branch_remote = partialmethod(_backend_func, "active_branch_remote")
    list_branches = partialmethod(_backend_func, "list_branches")
    list_tags = partialmethod(_backend_func, "list_tags")
    list_all_commits = partialmethod(_backend_func, "list_all_commits")
    get_rev = partialmethod(_backend_func, "get_rev")
    resolve_rev = partialmethod(_backend_func, "resolve_rev")
    resolve_commit = partialmethod(_backend_func, "resolve_commit")

    set_ref = partialmethod(_backend_func, "set_ref")
    get_ref = partialmethod(_backend_func, "get_ref")
    remove_ref = partialmethod(_backend_func, "remove_ref")
    iter_refs = partialmethod(_backend_func, "iter_refs")
    iter_remote_refs = partialmethod(_backend_func, "iter_remote_refs")
    get_refs_containing = partialmethod(_backend_func, "get_refs_containing")
    push_refspecs = partialmethod(_backend_func, "push_refspecs")
    _stash_iter = partialmethod(_backend_func, "_stash_iter")
    _stash_push = partialmethod(_backend_func, "_stash_push")
    _stash_apply = partialmethod(_backend_func, "_stash_apply")
    _stash_drop = partialmethod(_backend_func, "_stash_drop")
    _describe = partialmethod(_backend_func, "_describe")
    diff = partialmethod(_backend_func, "diff")
    reset = partialmethod(_backend_func, "reset")
    checkout_index = partialmethod(_backend_func, "checkout_index")
    status = partialmethod(_backend_func, "status")
    merge = partialmethod(_backend_func, "merge")
    validate_git_remote = partialmethod(_backend_func, "validate_git_remote")
    get_remote_url = partialmethod(_backend_func, "get_remote_url")
    check_ref_format = partialmethod(_backend_func, "check_ref_format")
    get_tag = partialmethod(_backend_func, "get_tag")
    get_config = partialmethod(_backend_func, "get_config")
    check_attr = partialmethod(_backend_func, "check_attr")

    get_tree_obj = partialmethod(_backend_func, "get_tree_obj")
    list_submodules = partialmethod(_backend_func, "list_submodules")

    def branch_revs(self, branch: str, end_rev: Optional[str] = None) -> Iterable[str]:
        """Iterate over revisions in a given branch (from newest to oldest).

        If end_rev is set, iterator will stop when the specified revision is
        reached.
        """
        commit = self.resolve_commit(branch)
        while commit is not None:
            yield commit.hexsha
            parent = first(commit.parents)
            if parent is None or parent == end_rev:
                return
            commit = self.resolve_commit(parent)

    @contextmanager
    def detach_head(
        self,
        rev: Optional[str] = None,
        force: bool = False,
        client: str = "scm",
    ):
        """Context manager for performing detached HEAD SCM operations.

        Detaches and restores HEAD similar to interactive git rebase.
        Restore is equivalent to 'reset --soft', meaning the caller is
        is responsible for preserving & restoring working tree state
        (i.e. via stash) when applicable.

        Yields revision of detached head.
        """
        if not rev:
            rev = "HEAD"
        orig_head = self.get_ref("HEAD", follow=False)
        logger.debug("Detaching HEAD at '%s'", rev)
        self.checkout(rev, detach=True, force=force)
        try:
            yield self.get_ref("HEAD")
        finally:
            prefix = self.LOCAL_BRANCH_PREFIX
            if orig_head.startswith(prefix):
                symbolic = True
                name = orig_head[len(prefix) :]
            else:
                symbolic = False
                name = orig_head
            self.set_ref(
                "HEAD",
                orig_head,
                symbolic=symbolic,
                message=f"{client}: Restore HEAD to '{name}'",
            )
            logger.debug("Restore HEAD to '%s'", name)
            self.reset()

    @contextmanager
    def stash_workspace(self, reinstate_index=False, **kwargs):
        """Stash and restore any workspace changes.

        Yields revision of the stash commit. Yields None if there were no
        changes to stash.
        """
        logger.debug("Stashing workspace")
        rev = self.stash.push(**kwargs)
        try:
            yield rev
        finally:
            if rev:
                logger.debug("Restoring stashed workspace")
                self.stash.pop(reinstate_index=reinstate_index)

    def _reset(self) -> None:
        self.backends.reset_all()

    def describe(
        self,
        revs: Iterable[str],
        base: Optional[str] = None,
        match: Optional[str] = None,
        exclude: Optional[str] = None,
    ) -> dict[str, Optional[str]]:
        results: dict[str, Optional[str]] = {}
        remained_revs = set()
        if base == "refs/heads":
            current_rev = self.get_rev()
            head_ref = self.get_ref("HEAD", follow=False)
            for rev in revs:
                if current_rev == rev and head_ref.startswith(base):
                    results[rev] = self.get_ref("HEAD", follow=False)
                else:
                    remained_revs.add(rev)
        else:
            remained_revs = set(revs)
        if remained_revs:
            results.update(self._describe(remained_revs, base, match, exclude))
        return results



import datetime
import stat
from abc import ABC, abstractmethod
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Optional, cast

from pygtrie import Trie

S_IFGITLINK = 0o160000


def S_ISGITLINK(m: int) -> bool:  # noqa: N802
    return stat.S_IFMT(m) == S_IFGITLINK


def _to_datetime(time: int, offset: int) -> datetime.datetime:
    tz = datetime.timezone(datetime.timedelta(seconds=offset))
    return datetime.datetime.fromtimestamp(time, tz=tz)


class GitObject(ABC):
    @abstractmethod
    def open(self, mode: str = "r", encoding: Optional[str] = None, **kwargs):
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def mode(self) -> int:
        pass

    @abstractmethod
    def scandir(self) -> Iterable["GitObject"]:
        pass

    @property
    def isfile(self) -> bool:
        return stat.S_ISREG(self.mode)

    @property
    def isdir(self) -> bool:
        return stat.S_ISDIR(self.mode)

    @property
    def issubmodule(self) -> bool:
        return S_ISGITLINK(self.mode)

    @property
    @abstractmethod
    def size(self) -> int:
        pass

    @property
    @abstractmethod
    def sha(self) -> str:
        pass


class GitTrie:
    def __init__(self, tree: GitObject, rev: str):
        self.tree = tree
        self.rev = rev
        self.trie = Trie()

        self.trie[()] = tree
        self._build(tree, ())

    def _build(self, tree: GitObject, path: tuple):
        for obj in tree.scandir():
            obj_path = (*path, obj.name)
            self.trie[obj_path] = obj

            if obj.isdir:
                self._build(obj, obj_path)

    def open(
        self,
        key: tuple,
        mode: Optional[str] = "r",
        encoding: Optional[str] = None,
        raw: bool = True,
    ):
        obj = self.trie[key]
        if obj.isdir:
            raise IsADirectoryError

        return obj.open(mode=mode, encoding=encoding, key=key, raw=raw, rev=self.rev)

    def exists(self, key: tuple) -> bool:
        return bool(self.trie.has_node(key))

    def isdir(self, key: tuple) -> bool:
        try:
            obj = self.trie[key]
        except KeyError:
            return False
        return obj.isdir

    def isfile(self, key: tuple) -> bool:
        try:
            obj = self.trie[key]
        except KeyError:
            return False

        return obj.isfile

    def ls(self, key: tuple):
        ret = []

        def node_factory(_, _key, children, obj):
            if key == _key:
                assert obj.isdir
                list(filter(None, children))
            else:
                ret.append(_key[-1])

        self.trie.traverse(node_factory, prefix=key)

        return ret

    def walk(self, top: tuple, topdown: Optional[bool] = True):
        dirs = []
        nondirs = []

        for name in self.ls(top):
            info = self.info((*top, name))
            if info["type"] == "directory":
                dirs.append(name)
            else:
                nondirs.append(name)

        if topdown:
            yield top, dirs, nondirs

        for dname in dirs:
            yield from self.walk((*top, dname), topdown=topdown)

        if not topdown:
            yield top, dirs, nondirs

    def info(self, key: tuple) -> dict:
        from scmrepo.utils import LazyDict

        obj = self.trie[key]

        def size():
            return obj.size

        ret = LazyDict(
            {
                "size": size,
                "type": "directory" if stat.S_ISDIR(obj.mode) else "file",
                "sha": obj.sha,
                "mode": obj.mode,
            }
        )

        return cast("dict", ret)


@dataclass
class GitCommit:
    hexsha: str
    commit_time: int
    commit_time_offset: int
    message: str
    parents: list[str]
    committer_name: str
    committer_email: str
    author_name: str
    author_email: str
    author_time: int
    author_time_offset: int

    @property
    def commit_datetime(self) -> datetime.datetime:
        return _to_datetime(self.commit_time, self.commit_time_offset)

    @property
    def author_datetime(self) -> datetime.datetime:
        return _to_datetime(self.author_time, self.author_time_offset)


@dataclass
class GitTag:
    name: str
    hexsha: str  # SHA for the tag object itself
    target: str  # SHA for the object the tag points to
    tagger_name: str
    tagger_email: str
    tag_time: int
    tag_time_offset: int
    message: str

    @property
    def tag_datetime(self) -> datetime.datetime:
        return _to_datetime(self.tag_time, self.tag_time_offset)



"""git stash convenience wrapper."""

import logging
from typing import Optional

from scmrepo.exceptions import SCMError

logger = logging.getLogger(__name__)


class Stash:
    """Wrapper for representing any arbitrary Git stash ref."""

    DEFAULT_STASH = "refs/stash"

    def __init__(self, scm, ref: Optional[str] = None):
        self.ref = ref if ref else self.DEFAULT_STASH
        self.scm = scm

    def __iter__(self):
        yield from self.scm._stash_iter(self.ref)

    def __len__(self):
        return len(self.list())

    def __getitem__(self, index):
        return self.list()[index]

    def list(self):
        return list(iter(self))

    def push(
        self, message: Optional[str] = None, include_untracked: bool = False
    ) -> Optional[str]:
        logger.debug("Stashing changes in '%s'", self.ref)
        rev, reset = self.scm._stash_push(  # pylint: disable=protected-access
            self.ref, message=message, include_untracked=include_untracked
        )
        if not rev:
            logger.debug("No changes to stash")
            return None
        if reset:
            self.scm.reset(hard=True)
        return rev

    def pop(self, **kwargs):
        """Pop the last stash commit.

        Supports the same keyword arguments as apply().
        """
        logger.debug("Popping from stash '%s'", self.ref)
        ref = f"{self.ref}@{{0}}"
        rev = self.scm.resolve_rev(ref)
        try:
            self.apply(rev, **kwargs)
        except Exception as exc:
            raise SCMError("Could not apply stash commit") from exc
        self.drop()
        return rev

    def apply(
        self,
        rev: str,
        reinstate_index: bool = False,
        skip_conflicts: bool = False,
    ):
        """Apply a stash commit.

        Arguments:
            rev: Stash commit to apply.
            reinstate_index: If True, stashed index changes will be reapplied.
            skip_conflicts: If True, conflicting changes will be skipped and
                will not be applied from the stash. By default, apply will
                fail if any conflicts are found.
        """
        logger.debug("Applying stash commit '%s'", rev)
        self.scm._stash_apply(  # pylint: disable=protected-access
            rev, reinstate_index=reinstate_index, skip_conflicts=skip_conflicts
        )

    def drop(self, index: int = 0):
        if index < 0 or index >= len(self):
            raise SCMError(f"Invalid stash ref '{self.ref}@{{{index}}}'")
        logger.debug("Dropping '%s@{%d}'", self.ref, index)
        self.scm._stash_drop(self.ref, index)  # pylint: disable=protected-access

    def clear(self):
        logger.debug("Clear stash '%s'", self.ref)
        for _ in range(len(self)):
            self.drop()



import os
from typing import Union

from .base import Base


# Syntactic sugar to signal that this is an actual implementation for a DVC
# project under no SCM control.
class NoSCM(Base):
    def __init__(
        self,
        root_dir: Union[str, os.PathLike[str], None] = None,
        _raise_not_implemented_as=NotImplementedError,
    ):
        super().__init__(root_dir=root_dir)
        self._exc = _raise_not_implemented_as

    def __getattr__(self, name):
        raise self._exc



import locale
import logging
import os
import stat
from collections.abc import Iterable, Iterator, Mapping
from contextlib import contextmanager
from io import BytesIO, StringIO, TextIOWrapper
from typing import (
    TYPE_CHECKING,
    Callable,
    Optional,
    Union,
)
from urllib.parse import urlparse

from funcy import cached_property, reraise

from scmrepo.exceptions import (
    CloneError,
    InvalidRemote,
    MergeConflictError,
    RevError,
    SCMError,
)
from scmrepo.git.backend.base import BaseGitBackend, SyncStatus
from scmrepo.git.config import Config
from scmrepo.git.objects import GitCommit, GitObject, GitTag
from scmrepo.urls import is_scp_style_url
from scmrepo.utils import relpath

logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from pygit2 import Commit, Oid, Signature
    from pygit2.config import Config as _Pygit2Config
    from pygit2.enums import CheckoutStrategy
    from pygit2.remotes import Remote
    from pygit2.repository import Repository

    from scmrepo.progress import GitProgressEvent


class Pygit2Object(GitObject):
    def __init__(self, obj, backend: Optional["Pygit2Backend"] = None):
        self.obj = obj
        self.backend = backend

    def open(
        self,
        mode: str = "r",
        encoding: Optional[str] = None,
        key: Optional[tuple[str, ...]] = None,
        raw: bool = True,
        rev: Optional[str] = None,
        **kwargs,
    ):
        from pygit2 import BlobIO, GitError

        if not encoding:
            encoding = locale.getpreferredencoding(False)
        if self.backend is not None:
            try:
                if rev:
                    # pylint: disable-next=protected-access
                    commit, _ref = self.backend._resolve_refish(rev)
                else:
                    pass
                if raw:
                    blob_kwargs = {}
                else:
                    assert key is not None
                    path = "/".join(key)
                    blob_kwargs = {
                        "as_path": path,
                        "commit_id": commit.id,
                    }
                blobio = BlobIO(self.obj, **blob_kwargs)
                if mode == "rb":
                    return blobio
                return TextIOWrapper(blobio, encoding=encoding)
            except GitError as exc:
                raise SCMError("failed to read git blob") from exc
        else:
            data = self.obj.read_raw()
        if mode == "rb":
            return BytesIO(data)
        return StringIO(data.decode(encoding))

    @property
    def name(self) -> str:
        return self.obj.name

    @property
    def mode(self):
        if not self.obj.filemode and self.obj.type_str == "tree":
            return stat.S_IFDIR
        return self.obj.filemode

    @cached_property
    def size(self) -> int:  # pylint: disable=invalid-overridden-method
        # NOTE: obj.size is currently only available for blobs
        if self.obj.type_str == "blob":
            return self.obj.size
        try:
            return len(self.obj.read_raw())
        except KeyError:
            return 0

    @property
    def sha(self) -> str:
        return str(self.obj.id)

    def scandir(self) -> Iterable["Pygit2Object"]:
        for entry in self.obj:
            yield Pygit2Object(entry, backend=self.backend)


class Pygit2Config(Config):
    def __init__(self, config: "_Pygit2Config"):
        self._config = config

    def _key(self, section: tuple[str, ...], name: str) -> str:
        return ".".join((*section, name))

    def get(self, section: tuple[str, ...], name: str) -> str:
        return self._config[self._key(section, name)]

    def get_bool(self, section: tuple[str, ...], name: str) -> bool:
        from pygit2 import GitError

        try:
            return self._config.get_bool(self._key(section, name))
        except GitError as exc:
            raise ValueError("invalid boolean config entry") from exc

    def get_multivar(self, section: tuple[str, ...], name: str) -> Iterator[str]:
        from pygit2 import GitError

        try:
            yield from self._config.get_multivar(self._key(section, name))
        except GitError as exc:
            raise ValueError("invalid multivar config entry") from exc


class Pygit2Backend(BaseGitBackend):  # pylint:disable=abstract-method
    def __init__(  # pylint:disable=W0231
        self, root_dir=os.curdir, search_parent_directories=True
    ):
        import pygit2

        from .filter import LFSFilter

        if search_parent_directories:
            ceiling_dirs = ""
        else:
            ceiling_dirs = os.path.abspath(root_dir)

        # NOTE: discover_repository will return path/.git/
        path = pygit2.discover_repository(  # pylint:disable=no-member
            os.fspath(root_dir), True, ceiling_dirs
        )
        if not path:
            raise SCMError(f"{root_dir} is not a git repository")

        self.repo = pygit2.Repository(path)

        self._stashes: dict = {}

        try:
            # NOTE: we want this init to be lazy so we do it on backend init.
            # for subsequent backend instances, this call will error out since
            # the filter is already registered
            pygit2.filter_register("lfs", LFSFilter)  # type: ignore[attr-defined]
        except ValueError:
            pass

    def close(self):
        if hasattr(self, "_refdb"):
            del self._refdb
        self.repo.free()

    @property
    def root_dir(self) -> Optional[str]:
        return self.repo.workdir

    @cached_property
    def _refdb(self):
        from pygit2 import RefdbFsBackend

        return RefdbFsBackend(self.repo)

    def _resolve_refish(self, refish: str):
        from pygit2 import Tag
        from pygit2.enums import ObjectType

        commit, ref = self.repo.resolve_refish(refish)  # type: ignore[attr-defined]
        if isinstance(commit, Tag):
            ref = commit
            commit = commit.peel(ObjectType.COMMIT)  # type: ignore[call-overload]
        return commit, ref

    @property
    def default_signature(self) -> "Signature":
        try:
            return self.repo.default_signature
        except KeyError as exc:
            raise SCMError("Git username and email must be configured") from exc

    @property
    def author(self) -> "Signature":
        return self._get_signature("GIT_AUTHOR")

    @property
    def committer(self) -> "Signature":
        return self._get_signature("GIT_COMMITTER")

    def _get_signature(self, name: str) -> "Signature":
        from pygit2 import Signature

        try:
            sig = self.default_signature
        except SCMError:
            logger.debug("No default signature (config is not set in repo)")
            sig = None

        if os.environ.get(f"{name}_DATE"):
            raise NotImplementedError("signature date override unsupported")

        user_name = os.environ.get(f"{name}_NAME", sig.name if sig else None)
        user_email = os.environ.get(f"{name}_EMAIL", sig.email if sig else None)

        if not user_email or not user_name:
            raise SCMError("Git user name and email must be configured")

        if sig:
            return Signature(
                name=user_name,
                email=user_email,
                time=sig.time,
                offset=sig.offset,
            )

        return Signature(
            name=user_name,
            email=user_email,
        )

    @staticmethod
    def _get_checkout_strategy(
        strategy: Optional["CheckoutStrategy"] = None,
    ) -> "CheckoutStrategy":
        from pygit2.enums import CheckoutStrategy

        if strategy is None:
            strategy = CheckoutStrategy.SAFE | CheckoutStrategy.RECREATE_MISSING
        if os.name == "nt":
            strategy |= CheckoutStrategy.SKIP_LOCKED_DIRECTORIES
        return strategy

    # Workaround to force git_backend_odb_pack to release open file handles
    # in DVC's mixed git-backend environment.
    # See https://github.com/iterative/dvc/issues/5641
    @contextmanager
    def release_odb_handles(self):
        yield
        # It is safe to free the libgit2 repo context multiple times - free
        # just forces libgit/pygit to release git ODB related contexts which
        # can be reacquired later as needed.
        self.repo.free()

    def list_submodules(self) -> list[str]:
        return self.repo.listall_submodules()

    @classmethod
    def clone(
        cls,
        url: str,
        to_path: Union[str, os.PathLike[str]],
        shallow_branch: Optional[str] = None,
        progress: Optional[Callable[["GitProgressEvent"], None]] = None,
        bare: bool = False,
        mirror: bool = False,
    ):
        from pygit2 import GitError, clone_repository

        from .callbacks import RemoteCallbacks

        if shallow_branch:
            raise NotImplementedError
        if mirror:
            bare = True
        try:
            with RemoteCallbacks(progress=progress) as cb:
                repo = clone_repository(
                    url, os.fspath(to_path), callbacks=cb, bare=bare
                )
                if mirror:
                    cls._set_mirror(repo, progress=progress)
        except GitError as exc:
            raise CloneError(url, os.fsdecode(to_path)) from exc

    @staticmethod
    def _set_mirror(
        repo: "Repository",
        progress: Optional[Callable[["GitProgressEvent"], None]] = None,
    ):
        from .callbacks import RemoteCallbacks

        url = repo.remotes["origin"].url
        repo.remotes.delete("origin")
        # NOTE: Pygit2 remotes.create("origin", url, fetch_refspec) creates a
        # duplicate config section for each remote config entry. We just edit
        # the config directly so that it creates a single section to be
        # consistent with CLI Git
        repo.config["remote.origin.url"] = url
        repo.config["remote.origin.fetch"] = "+refs/*:refs/*"
        repo.config["remote.origin.mirror"] = True
        with RemoteCallbacks(progress=progress) as cb:
            repo.remotes["origin"].fetch(callbacks=cb)

    @staticmethod
    def init(path: str, bare: bool = False) -> None:
        from pygit2 import init_repository

        init_repository(path, bare=bare)

    @property
    def dir(self) -> str:
        raise NotImplementedError

    def add(
        self,
        paths: Union[str, Iterable[str]],
        update: bool = False,
        force: bool = False,
    ):
        raise NotImplementedError

    def commit(self, msg: str, no_verify: bool = False):
        raise NotImplementedError

    def checkout(
        self,
        branch: str,
        create_new: Optional[bool] = False,
        force: bool = False,
        **kwargs,
    ):
        from pygit2 import GitError
        from pygit2.enums import CheckoutStrategy

        strategy = self._get_checkout_strategy(
            CheckoutStrategy.FORCE if force else None
        )

        with self.release_odb_handles():
            if create_new:
                commit = self.repo.revparse_single("HEAD")
                new_branch = self.repo.branches.local.create(branch, commit)  # type: ignore[arg-type]
                self.repo.checkout(new_branch, strategy=strategy)  # type: ignore[attr-defined]
            else:
                if branch == "-":
                    branch = "@{-1}"
                try:
                    commit, ref = self._resolve_refish(branch)
                except (KeyError, GitError) as exc:
                    raise RevError(f"unknown Git revision '{branch}'") from exc
                self.repo.checkout_tree(commit, strategy=strategy)  # type: ignore[attr-defined]
                detach = kwargs.get("detach", False)
                if ref and not detach:
                    self.repo.set_head(ref.name)  # type: ignore[attr-defined]
                else:
                    self.repo.set_head(commit.id)  # type: ignore[attr-defined]

    def fetch(
        self,
        remote: Optional[str] = None,
        force: bool = False,
        unshallow: bool = False,
    ):
        raise NotImplementedError

    def pull(self, **kwargs):
        raise NotImplementedError

    def push(self):
        raise NotImplementedError

    def branch(self, branch: str):
        from pygit2 import GitError

        try:
            commit = self.repo[self.repo.head.target]
            self.repo.create_branch(branch, commit)  # type: ignore[arg-type]
        except GitError as exc:
            raise SCMError(f"Failed to create branch '{branch}'") from exc

    def tag(
        self,
        tag: str,
        target: Optional[str] = None,
        annotated: bool = False,
        message: Optional[str] = None,
    ):
        from pygit2 import GitError
        from pygit2.enums import ObjectType

        if annotated and not message:
            raise SCMError("message is required for annotated tag")
        target_obj = self.repo.revparse_single(target or "HEAD")
        with reraise(GitError, SCMError("Failed to create tag")):
            self.repo.create_tag(
                tag,
                target_obj.id,
                ObjectType.COMMIT,
                self.committer,
                message or "",
            )

    def untracked_files(self) -> Iterable[str]:
        raise NotImplementedError

    def is_tracked(self, path: str) -> bool:
        raise NotImplementedError

    def is_dirty(self, untracked_files: bool = False) -> bool:
        raise NotImplementedError

    def active_branch(self) -> str:
        if self.repo.head_is_detached:
            raise SCMError("No active branch (detached HEAD)")
        if self.repo.head_is_unborn:
            # if HEAD points to a nonexistent branch we still return the
            # branch name (without "refs/heads/" prefix) to match gitpython's
            # behavior
            return self.repo.references["HEAD"].target[11:]  # type: ignore[index]
        return self.repo.head.shorthand

    def active_branch_remote(self) -> str:
        try:
            upstream = self.repo.branches[self.active_branch()].upstream
            if upstream:
                return upstream.remote_name
        except (KeyError, ValueError):
            pass
        raise SCMError("No active branch tracking remote")

    def list_branches(self) -> Iterable[str]:
        base = "refs/heads/"
        return sorted(ref[len(base) :] for ref in self.iter_refs(base))

    def list_tags(self) -> Iterable[str]:
        base = "refs/tags/"
        return sorted(ref[len(base) :] for ref in self.iter_refs(base))

    def list_all_commits(self) -> Iterable[str]:
        import pygit2
        from pygit2.enums import SortMode

        # Add HEAD
        starting_points: list[Union[Oid, str]] = []
        if not self.repo.head_is_unborn:
            starting_points.append(self.repo.head.target)

        # Add all branches, remotes, and tags
        for ref in self.repo.references:
            if ref.startswith(("refs/heads/", "refs/remotes/")):
                oid = self.repo.revparse_single(ref).id
                starting_points.append(oid)
            elif ref.startswith("refs/tags/"):
                tag_obj = self.repo.revparse_single(ref)
                if isinstance(tag_obj, pygit2.Tag):
                    starting_points.append(tag_obj.target)
                else:
                    starting_points.append(tag_obj.id)

        # Walk all commits
        walker = self.repo.walk(None)
        for o in starting_points:
            walker.push(o)
        walker.sort(SortMode.TIME)
        return [str(commit.id) for commit in walker]

    def get_tree_obj(self, rev: str, **kwargs) -> Pygit2Object:
        tree = self.repo[rev].tree  # type: ignore[attr-defined]
        return Pygit2Object(tree, backend=self)

    def get_rev(self) -> str:
        raise NotImplementedError

    def resolve_rev(self, rev: str) -> str:
        from pygit2 import GitError

        try:
            commit, _ref = self._resolve_refish(rev)
            return str(commit.id)
        except (KeyError, GitError):
            pass

        # Look for single exact match in remote refs
        shas = {
            self.get_ref(f"refs/remotes/{remote.name}/{rev}")
            for remote in self.repo.remotes
        } - {None}
        if len(shas) > 1:
            raise RevError(f"ambiguous Git revision '{rev}'")
        if len(shas) == 1:
            sha = shas.pop()
            if sha is not None:
                return sha
        raise RevError(f"unknown Git revision '{rev}'")

    def resolve_commit(self, rev: str) -> "GitCommit":
        from pygit2 import GitError

        try:
            commit, _ref = self._resolve_refish(rev)
        except (KeyError, GitError) as exc:
            raise SCMError(f"Invalid commit '{rev}'") from exc
        return GitCommit(
            str(commit.id),
            commit.commit_time,
            commit.commit_time_offset * 60,
            commit.message,
            [str(parent) for parent in commit.parent_ids],
            commit.committer.name,
            commit.committer.email,
            commit.author.name,
            commit.author.email,
            commit.author.time,
            commit.author.offset * 60,
        )

    def _get_stash(self, ref: str):
        raise NotImplementedError

    def is_ignored(self, path: "Union[str, os.PathLike[str]]") -> bool:
        rel = relpath(path, self.root_dir)
        if os.name == "nt":
            rel = rel.replace("\\", "/")
        return self.repo.path_is_ignored(rel)

    def set_ref(
        self,
        name: str,
        new_ref: str,
        old_ref: Optional[str] = None,
        message: Optional[str] = None,
        symbolic: Optional[bool] = False,
    ):
        if old_ref and old_ref != self.get_ref(name, follow=False):
            raise SCMError(f"Failed to set '{name}'")

        if message:
            self._refdb.ensure_log(name)
        if symbolic:
            self.repo.create_reference_symbolic(name, new_ref, True, message=message)
        else:
            self.repo.create_reference_direct(name, new_ref, True, message=message)

    def get_ref(self, name, follow: bool = True) -> Optional[str]:
        from pygit2 import InvalidSpecError, Tag
        from pygit2.enums import ObjectType, ReferenceType

        try:
            ref = self.repo.references.get(name)
        except InvalidSpecError:
            return None
        if not ref:
            return None
        if follow and ref.type == ReferenceType.SYMBOLIC:
            ref = ref.resolve()
        try:
            obj = self.repo[ref.target]
            if isinstance(obj, Tag):
                return str(obj.peel(ObjectType.COMMIT).id)  # type: ignore[call-overload]
        except ValueError:
            pass

        return str(ref.target)

    def remove_ref(self, name: str, old_ref: Optional[str] = None):
        ref = self.repo.references.get(name)
        if not ref and not old_ref:
            return
        if old_ref and old_ref != str(ref.target):
            raise SCMError(f"Failed to remove '{name}'")
        ref.delete()

    def iter_refs(self, base: Optional[str] = None):
        if base:
            for ref in self.repo.references:
                if ref.startswith(base):
                    yield ref
        else:
            yield from self.repo.references

    def get_refs_containing(self, rev: str, pattern: Optional[str] = None):
        import fnmatch

        from pygit2 import GitError

        def _contains(repo, ref, search_commit):
            commit, _ref = self._resolve_refish(ref)
            base = repo.merge_base(search_commit.id, commit.id)
            return base == search_commit.id

        try:
            search_commit, _ref = self._resolve_refish(rev)
        except (KeyError, GitError) as exc:
            raise SCMError(f"Invalid rev '{rev}'") from exc

        if not pattern:
            yield from (
                ref
                for ref in self.iter_refs()
                if _contains(self.repo, ref, search_commit)
            )
            return

        literal = pattern.rstrip("/").split("/")
        for ref in self.iter_refs():
            if (
                ref.split("/")[: len(literal)] == literal
                or fnmatch.fnmatch(ref, pattern)
            ) and _contains(self.repo, ref, search_commit):
                yield ref

    def push_refspecs(
        self,
        url: str,
        refspecs: Union[str, Iterable[str]],
        force: bool = False,
        on_diverged: Optional[Callable[[str, str], bool]] = None,
        progress: Optional[Callable[["GitProgressEvent"], None]] = None,
        **kwargs,
    ) -> Mapping[str, SyncStatus]:
        raise NotImplementedError

    def _merge_remote_branch(
        self,
        rh: str,
        lh: str,
        force: bool = False,
        on_diverged: Optional[Callable[[str, str], bool]] = None,
    ) -> SyncStatus:
        from pygit2.enums import MergeAnalysis

        rh_rev = self.resolve_rev(rh)

        if force:
            self.set_ref(lh, rh_rev)
            return SyncStatus.SUCCESS

        try:
            merge_result, _ = self.repo.merge_analysis(rh_rev, lh)
        except KeyError:
            self.set_ref(lh, rh_rev)
            return SyncStatus.SUCCESS

        if merge_result & MergeAnalysis.UP_TO_DATE:
            return SyncStatus.UP_TO_DATE
        if merge_result & MergeAnalysis.FASTFORWARD:
            self.set_ref(lh, rh_rev)
            return SyncStatus.SUCCESS
        if merge_result & MergeAnalysis.NORMAL:
            if on_diverged and on_diverged(lh, rh_rev):
                return SyncStatus.SUCCESS
            return SyncStatus.DIVERGED
        logger.debug("Unexpected merge result: %s", MergeAnalysis.NORMAL)
        raise SCMError("Unknown merge analysis result")

    @contextmanager
    def _get_remote(self, url: str) -> Iterator["Remote"]:
        """Return a pygit2.Remote suitable for the specified Git URL or remote name."""
        try:
            remote = self.repo.remotes[url]
            url = remote.url
        except ValueError:
            pass
        except KeyError as exc:
            raise SCMError(f"'{url}' is not a valid Git remote or URL") from exc

        if os.name == "nt":
            url = url.removeprefix("file://")
        remote = self.repo.remotes.create_anonymous(url)
        parsed = urlparse(remote.url)
        if parsed.scheme in ("git", "git+ssh", "ssh") or is_scp_style_url(remote.url):
            raise NotImplementedError
        yield remote

    def fetch_refspecs(
        self,
        url: str,
        refspecs: Union[str, Iterable[str]],
        force: bool = False,
        on_diverged: Optional[Callable[[str, str], bool]] = None,
        progress: Optional[Callable[["GitProgressEvent"], None]] = None,
        **kwargs,
    ) -> Mapping[str, SyncStatus]:
        import fnmatch

        from pygit2 import GitError

        from .callbacks import RemoteCallbacks

        refspecs = self._refspecs_list(refspecs, force=force)

        # libgit2 rejects diverged refs but does not have a callback to notify
        # when a ref was rejected so we have to determine whether no callback
        # means up to date or rejected
        def _default_status(
            src: str, dst: str, remote_refs: dict[str, "Oid"]
        ) -> SyncStatus:
            try:
                if remote_refs[src] != self.repo.references[dst].target:
                    return SyncStatus.DIVERGED
            except KeyError:
                # remote_refs lookup is skipped when force is set, refs cannot
                # be diverged on force
                pass
            return SyncStatus.UP_TO_DATE

        with self._get_remote(url) as remote:
            with reraise(
                GitError,
                SCMError(f"Git failed to fetch ref from '{url}'"),
            ):
                with RemoteCallbacks(progress=progress) as cb:
                    remote_refs: dict[str, Oid] = (
                        {
                            head["name"]: head["oid"]
                            for head in remote.ls_remotes(callbacks=cb, proxy=True)
                        }
                        if not force
                        else {}
                    )
                    remote.fetch(
                        refspecs=refspecs, callbacks=cb, message="fetch", proxy=True
                    )

            result: dict[str, SyncStatus] = {}
            for refspec in refspecs:
                lh, rh = refspec.split(":")
                if lh.endswith("*"):
                    assert rh.endswith("*")
                    lh_prefix = lh[:-1]
                    rh_prefix = rh[:-1]
                    for refname in remote_refs:
                        if fnmatch.fnmatch(refname, lh):
                            src = refname
                            dst = f"{rh_prefix}{refname[len(lh_prefix) :]}"
                            result[dst] = cb.result.get(
                                src, _default_status(src, dst, remote_refs)
                            )
                else:
                    result[rh] = cb.result.get(lh, _default_status(lh, rh, remote_refs))

        return result

    @staticmethod
    def _refspecs_list(
        refspecs: Union[str, Iterable[str]],
        force: bool = False,
    ) -> list[str]:
        if isinstance(refspecs, str):
            if force and not refspecs.startswith("+"):
                refspecs = f"+{refspecs}"
            return [refspecs]
        if force:
            return [
                (refspec if refspec.startswith("+") else f"+{refspec}")
                for refspec in refspecs
            ]
        return list(refspecs)

    def _stash_iter(self, ref: str):
        raise NotImplementedError

    def _stash_push(
        self,
        ref: str,
        message: Optional[str] = None,
        include_untracked: bool = False,
    ) -> tuple[Optional[str], bool]:
        from scmrepo.git import Stash

        try:
            oid = self.repo.stash(  # type: ignore[attr-defined]
                self.committer,
                message=message,
                include_untracked=include_untracked,
            )
        except KeyError:
            # GIT_ENOTFOUND, nothing to stash
            return None, False
        commit = self.repo[oid]

        if ref != Stash.DEFAULT_STASH:
            self.set_ref(ref, commit.id, message=commit.message)  # type: ignore[attr-defined,arg-type]
            self.repo.stash_drop()  # type: ignore[attr-defined]
        return str(oid), False

    def _stash_apply(
        self,
        rev: str,
        reinstate_index: bool = False,
        skip_conflicts: bool = False,
        **kwargs,
    ):
        from pygit2 import GitError
        from pygit2.enums import CheckoutStrategy

        from scmrepo.git import Stash

        def _apply(index):
            try:
                self.repo.index.read(False)  # type: ignore[attr-defined]
                strategy = self._get_checkout_strategy()
                if skip_conflicts:
                    strategy |= CheckoutStrategy.ALLOW_CONFLICTS
                self.repo.stash_apply(  # type: ignore[attr-defined]
                    index, strategy=strategy, reinstate_index=reinstate_index
                )
            except GitError as exc:
                raise MergeConflictError(
                    "Stash apply resulted in merge conflicts"
                ) from exc

        # libgit2 stash apply only accepts refs/stash items by index. If rev is
        # not in refs/stash, we will push it onto the stash, and then pop it
        commit, _ref = self._resolve_refish(rev)
        stash = self.repo.references.get(Stash.DEFAULT_STASH)
        if stash:
            for i, entry in enumerate(stash.log()):
                if entry.oid_new == commit.id:
                    _apply(i)
                    return

        self.set_ref(Stash.DEFAULT_STASH, commit.id, message=commit.message)
        try:
            _apply(0)
        finally:
            self.repo.stash_drop()  # type: ignore[attr-defined]

    def _stash_drop(self, ref: str, index: int):
        from scmrepo.git import Stash

        if ref != Stash.DEFAULT_STASH:
            raise NotImplementedError

        self.repo.stash_drop(index)  # type: ignore[attr-defined]

    def _describe(
        self,
        revs: Iterable[str],
        base: Optional[str] = None,
        match: Optional[str] = None,
        exclude: Optional[str] = None,
    ) -> Mapping[str, Optional[str]]:
        raise NotImplementedError

    def diff(self, rev_a: str, rev_b: str, binary=False) -> str:
        raise NotImplementedError

    def reset(self, hard: bool = False, paths: Optional[Iterable[str]] = None):
        from pygit2 import IndexEntry
        from pygit2.enums import ResetMode

        if self.list_submodules():
            raise NotImplementedError(
                "pygit2 seems to remove files from submodules on reset"
            )

        self.repo.index.read(False)  # type: ignore[attr-defined]
        if paths is not None:
            tree = self.repo.revparse_single("HEAD").tree  # type: ignore[attr-defined]
            for path in paths:
                rel = relpath(path, self.root_dir)
                if os.name == "nt":
                    rel = rel.replace("\\", "/")
                obj = tree[rel]
                self.repo.index.add(IndexEntry(rel, obj.id, obj.filemode))  # type: ignore[attr-defined]
            self.repo.index.write()  # type: ignore[attr-defined]
        elif hard:
            self.repo.reset(self.repo.head.target, ResetMode.HARD)
        else:
            self.repo.reset(self.repo.head.target, ResetMode.MIXED)

    def checkout_index(
        self,
        paths: Optional[Iterable[str]] = None,
        force: bool = False,
        ours: bool = False,
        theirs: bool = False,
    ):
        from pygit2.enums import CheckoutStrategy

        assert not (ours and theirs)
        strategy = CheckoutStrategy.RECREATE_MISSING
        if force or ours or theirs:
            strategy |= CheckoutStrategy.FORCE
        else:
            strategy |= CheckoutStrategy.SAFE

        if ours or theirs:
            strategy |= CheckoutStrategy.ALLOW_CONFLICTS
        strategy = self._get_checkout_strategy(strategy)

        index = self.repo.index  # type: ignore[attr-defined]
        if paths:
            path_list: Optional[list[str]] = [
                relpath(path, self.root_dir) for path in paths
            ]
            if os.name == "nt":
                path_list = [
                    path.replace("\\", "/")
                    for path in path_list  # type: ignore[union-attr]
                ]
        else:
            path_list = None

        with self.release_odb_handles():
            self.repo.checkout_index(index=index, paths=path_list, strategy=strategy)  # type: ignore[attr-defined]

            if index.conflicts and (ours or theirs):
                for ancestor, ours_entry, theirs_entry in index.conflicts:
                    if not ancestor:
                        continue
                    if ours:
                        entry = ours_entry
                        index.add(ours_entry)
                    else:
                        entry = theirs_entry
                    assert self.root_dir
                    path = os.path.join(self.root_dir, entry.path)
                    with open(path, "wb") as fobj:
                        fobj.write(self.repo.get(entry.id).read_raw())  # type: ignore[attr-defined]
                    index.add(entry.path)
                index.write()

    def status(
        self, ignored: bool = False, untracked_files: str = "all"
    ) -> tuple[Mapping[str, Iterable[str]], Iterable[str], Iterable[str]]:
        from pygit2.enums import FileStatus

        staged: Mapping[str, list[str]] = {
            "add": [],
            "delete": [],
            "modify": [],
        }
        unstaged: list[str] = []
        untracked: list[str] = []

        states = {
            FileStatus.WT_NEW: untracked,
            FileStatus.WT_MODIFIED: unstaged,
            FileStatus.WT_TYPECHANGE: staged["modify"],
            FileStatus.WT_DELETED: staged["modify"],
            FileStatus.WT_RENAMED: staged["modify"],
            FileStatus.INDEX_NEW: staged["add"],
            FileStatus.INDEX_MODIFIED: staged["modify"],
            FileStatus.INDEX_DELETED: staged["delete"],
            FileStatus.WT_UNREADABLE: untracked,
        }

        if untracked_files != "no" and ignored:
            states[FileStatus.IGNORED] = untracked

        for file, state in self.repo.status(
            untracked_files=untracked_files, ignored=ignored
        ).items():
            for git_state in states:
                flag = state & git_state
                if flag:
                    states[flag].append(file)

        return (
            {status: paths for status, paths in staged.items() if paths},
            unstaged,
            untracked,
        )

    def iter_remote_refs(self, url: str, base: Optional[str] = None, **kwargs):
        raise NotImplementedError

    def merge(  # noqa: C901
        self,
        rev: str,
        commit: bool = True,
        msg: Optional[str] = None,
        squash: bool = False,
    ) -> Optional[str]:
        from pygit2 import GitError
        from pygit2.enums import MergeAnalysis, MergePreference

        if commit and squash:
            raise SCMError("Cannot merge with 'squash' and 'commit'")

        with self.release_odb_handles():
            self.repo.index.read(False)  # type: ignore[attr-defined]
            obj, _ref = self.repo.resolve_refish(rev)  # type: ignore[attr-defined]
            try:
                analysis, ff_pref = self.repo.merge_analysis(obj.id)
            except GitError as exc:
                raise SCMError("Merge analysis failed") from exc

            if analysis == MergeAnalysis.NONE:
                raise SCMError(f"'{rev}' cannot be merged into HEAD")
            if analysis & MergeAnalysis.UP_TO_DATE:
                return None

            try:
                self.repo.merge(obj.id)  # type: ignore[attr-defined]
                self.repo.index.write()  # type: ignore[attr-defined]
            except GitError as exc:
                raise SCMError("Merge failed") from exc

            if self.repo.index.conflicts:  # type: ignore[attr-defined]
                raise MergeConflictError("Merge contained conflicts")

            try:
                if not (squash or ff_pref & MergePreference.NO_FASTFORWARD):
                    if analysis & MergeAnalysis.FASTFORWARD:
                        return self._merge_ff(rev, obj)

                    if analysis & MergeAnalysis.UNBORN:
                        self.repo.set_head(obj.id)  # type: ignore[attr-defined]
                        return str(obj.id)

                if ff_pref & MergePreference.FASTFORWARD_ONLY:
                    raise SCMError(f"Cannot fast-forward HEAD to '{rev}'")

                if commit:
                    return self._merge_commit(msg, obj)

                # --squash merge:
                # HEAD is not moved and merge changes stay in index
                return None
            finally:
                self.repo.state_cleanup()  # type: ignore[attr-defined]
                self.repo.index.write()  # type: ignore[attr-defined]

    def _merge_ff(self, rev: str, obj) -> str:
        if self.repo.head_is_detached:
            self.repo.set_head(obj.id)  # type: ignore[attr-defined]
        else:
            branch = self.get_ref("HEAD", follow=False)
            assert branch
            self.set_ref(
                branch,
                str(obj.id),
                message=f"merge {rev}: Fast-forward",
            )
        return str(obj.id)

    def _merge_commit(self, msg: Optional[str], obj) -> str:
        if not msg:
            raise SCMError("Merge commit message is required")
        tree = self.repo.index.write_tree()  # type: ignore[attr-defined]
        merge_commit = self.repo.create_commit(
            "HEAD",
            self.author,
            self.committer,
            msg,
            tree,
            [self.repo.head.target, obj.id],
        )
        return str(merge_commit)

    def validate_git_remote(self, url: str, **kwargs):
        raise NotImplementedError

    def get_remote_url(self, remote: str) -> str:
        try:
            return self.repo.remotes[remote].url
        except KeyError as exc:
            raise InvalidRemote(remote) from exc

    def check_ref_format(self, refname: str):
        raise NotImplementedError

    def get_tag(self, name: str) -> Optional[Union[str, "GitTag"]]:
        from pygit2 import InvalidSpecError, Tag

        try:
            ref = self.repo.references.get(f"refs/tags/{name}")
        except InvalidSpecError:
            return None
        if not ref:
            return None
        try:
            tag = self.repo[ref.target]
            if isinstance(tag, Tag):
                return GitTag(
                    tag.name,
                    str(tag.id),
                    str(tag.target),
                    tag.tagger.name,
                    tag.tagger.email,
                    tag.tagger.time,
                    tag.tagger.offset * 60,
                    tag.message,
                )
        except KeyError:
            pass
        return str(ref.target)

    def get_config(self, path: Optional[str] = None) -> "Config":
        from pygit2.config import Config as _Pygit2Config

        if path:
            return Pygit2Config(_Pygit2Config(path))
        return Pygit2Config(self.repo.config)  # type: ignore[attr-defined]

    def check_attr(
        self,
        path: str,
        attr: str,
        source: Optional[str] = None,
    ) -> Optional[Union[bool, str]]:
        from pygit2 import GitError
        from pygit2.enums import AttrCheck

        commit: Optional[Commit] = None
        flags = AttrCheck.FILE_THEN_INDEX
        if source:
            try:
                commit, _ref = self._resolve_refish(source)
                flags = AttrCheck.INDEX_ONLY | AttrCheck.INCLUDE_COMMIT
            except (KeyError, GitError) as exc:
                raise SCMError(f"Invalid commit '{source}'") from exc
        try:
            return self.repo.get_attr(  # type: ignore[attr-defined]
                path, attr, flags=flags, commit=commit.id if commit else None
            )
        except GitError as exc:
            raise SCMError("Failed to check attribute") from exc
